# -*- coding: utf-8 -*-
"""
Module scenario.py
===========================================

Module for sensitivity analysis in *PyDASA*.

This module provides the Sensitivity class for performing sensitivity analysis on dimensional coefficients derived from dimensional analysis.

Classes:

    **Sensitivity**: Performs sensitivity analysis on dimensional coefficients in *PyDASA*.

*IMPORTANT:* Based on the theory from:
    # H.Gorter, *Dimensionalanalyse: Eine Theoririe der physikalischen Dimensionen mit Anwendungen*
"""

from __future__ import annotations
from dataclasses import dataclass, field, fields
from typing import Optional, List, Dict, Any, Callable, Union

# Third-party modules
import numpy as np
import sympy as sp
from sympy import diff, lambdify    # , Symbol, symbols

from SALib.sample.fast_sampler import sample
from SALib.analyze.fast import analyze

# Import validation base classes
from pydasa.core.basic import Foundation

# Import validation decorators
from pydasa.validations.decorators import validate_type
from pydasa.validations.decorators import validate_choices
from pydasa.validations.decorators import validate_pattern
from pydasa.validations.decorators import validate_custom
from pydasa.validations.decorators import validate_dict_types
from pydasa.validations.decorators import validate_emptiness
from pydasa.validations.decorators import validate_index

# Import related classes
from pydasa.dimensional.vaschy import Schema
from pydasa.dimensional.buckingham import Coefficient
from pydasa.elements.parameter import Variable
from pydasa.core.setup import AnaliticMode

# Import utils
from pydasa.serialization.parser import parse_latex
from pydasa.serialization.parser import create_latex_mapping
from pydasa.serialization.parser import latex_to_python

# Import configuration
from pydasa.core.setup import PYDASA_CFG
from pydasa.validations.patterns import LATEX_RE


@dataclass
class Sensitivity(Foundation):
    # FIXME clean code, some vars and types are inconsistent
    """**Sensitivity** class for analyzing variable impacts in *PyDASA*.

    Performs sensitivity analysis on dimensionless coefficients to determine which variables have the most significant impact on the system behavior.

    Args:
        Foundation: Foundation class for validation of symbols and frameworks.

    Attributes:
        # Identification and Classification
        name (str): User-friendly name of the sensitivity analysis.
        description (str): Brief summary of the sensitivity analysis.
        _idx (int): Index/precedence of the sensitivity analysis.
        _sym (str): Symbol representation (LaTeX or alphanumeric).
        _alias (str): Python-compatible alias for use in code.
        _fwk (str): Frameworks context (PHYSICAL, COMPUTATION, SOFTWARE, CUSTOM).
        _cat (str): Category of analysis (SYM, NUM).

        # Expression Management
        _pi_expr (str): LaTeX expression to analyze.
        _sym_func (Callable): Sympy function of the sensitivity.
        _exe_func (Callable): Executable function for numerical evaluation.
        _variables (Dict[str, Variable]): Variable symbols in the expression.
        _var_names (List[str]): List of variable names extracted from the expression.
        _schema (Schema): Dimensional framework schema.
        _coefficient (Coefficient): Specifications of the dimensionless coefficient to analyse.
        _symbols (Dict[str, Any]): Python symbols for the variables.
        _aliases (Dict[str, Any]): Variable aliases for use in code.

        # Analysis Configuration
        _var_bounds (List[List[float]]): Min/max bounds for each variable.
        _var_values (Dict[str, float]): Values for symbolic analysis.
        _var_ranges (np.ndarray): Sample value range for numerical analysis.
        _var_domains (np.ndarray): Sample domain for numerical analysis.
        _experiments (int): Number of samples for analysis.

        # Results
        _results (Dict[str, Any]): Analysis results.
    """

    # Category attribute
    # :attr: _cat
    _cat: str = AnaliticMode.SYM.value
    """Category of sensitivity analysis (SYM, NUM)."""

    # Expression properties
    # :attr: _pi_expr
    _pi_expr: Optional[str] = None
    """LaTeX expression to analyze."""

    # :attr: _sym_func
    _sym_func: Optional[sp.Expr] = None
    """Sympy expression of the sensitivity (symbolic form)."""

    # :attr: _exe_func
    _exe_func: Optional[Union[Callable, Dict[str, Callable]]] = None
    """Executable function(s) for numerical evaluation (compiled from symbolic form)."""

    # :attr: _variables
    _variables: Dict[str, Variable] = field(default_factory=dict)
    """Variable symbols in the expression."""

    # :attr: _var_names
    _var_names: List[str] = field(default_factory=list)
    """List of variable names extracted from the expression (ordered)."""

    # :attr: _schema
    _schema: Schema = field(default_factory=Schema)
    """Dimensional framework schema for FDU management."""

    # :attr: _coefficient
    _coefficient: Optional[Coefficient] = None
    """Specifications of the dimensionless coefficient to analyse."""

    # :attr: _symbols
    _symbols: Dict[sp.Symbol, sp.Symbol] = field(default_factory=dict)
    """Python symbols for the variables."""

    # :attr: _aliases
    _aliases: Dict[str, sp.Symbol] = field(default_factory=dict)
    """Variable aliases for use in code."""

    # :attr: _latex_to_py
    _latex_to_py: Dict[str, str] = field(default_factory=dict)
    """Mapping from LaTeX symbols to Python-compatible names."""

    # :attr: _py_to_latex
    _py_to_latex: Dict[str, str] = field(default_factory=dict)
    """Mapping from Python-compatible names to LaTeX symbols."""

    # Analysis configuration
    # :attr: _var_bounds
    _var_bounds: List[List[float]] = field(default_factory=list)
    """Min/max bounds for each variable."""

    # :attr: _var_values
    _var_values: Dict[str, float] = field(default_factory=dict)
    """Values for symbolic analysis."""

    # :attr: _var_domains
    _var_domains: Optional[np.ndarray] = None
    """Sample domain (inputs) for numerical analysis."""

    # :attr: _var_ranges
    _var_ranges: Optional[np.ndarray] = None
    """Sample value range (results) for numerical analysis."""

    # :attr: _experiments
    _experiments: int = -1
    """Number of experiments/samples for analysis."""

    # Results
    # :attr: _results
    _results: Dict[str, Any] = field(default_factory=dict)
    """Analysis results."""

    def __post_init__(self) -> None:
        """*__post_init__()* Initializes the sensitivity analysis. Validates basic properties, sets default values, and processes the expression if provided.
        """
        # Initialize from base class
        super().__post_init__()

        # Set default symbol if not specified
        if not self._sym:
            self._sym = f"SENS_{{\\Pi_{{{self._idx}}}}}" if self._idx >= 0 else "SENS_{\\Pi_{-1}}"
        # Set default Python alias if not specified
        if not self._alias:
            self._alias = latex_to_python(self._sym)

        # Set name and description if not already set
        if not self.name:
            self.name = f"{self._sym} Sensitivity"
        if not self.description:
            self.description = f"Sensitivity analysis for {self._sym}"

        if self._pi_expr:
            # Parse the expression
            self._parse_expression(self._pi_expr)

    def _validate_sym_expr(self, val: Any) -> None:
        """*_validate_sym_expr()* Check that the symbolic expression is a SymPy expression.

        Args:
            val (Any): Value to validate.

        Raises:
            ValueError: If value is not a valid SymPy expression.
        """
        if not isinstance(val, sp.Expr):
            _msg = "'sym_func' must be a SymPy expression, "
            _msg += f"got {type(val).__name__}"
            raise ValueError(_msg)

    def _validate_exec_func(self, val: Any) -> None:
        """Validate executable function is callable or dict of callables."""
        if callable(val):
            return
        if isinstance(val, dict):
            if not all(callable(v) for v in val.values()):
                raise ValueError("All dictionary values must be callable")
            return
        _msg = "'exe_func' must be Callable or Dict[str, Callable], "
        _msg += f"got {type(val).__name__}"
        raise ValueError(_msg)

    def _validate_analysis_ready(self) -> None:
        """*_validate_analysis_ready()* Checks if the analysis can be performed.

        Raises:
            ValueError: If the variables are missing.
            ValueError: If the python-compatible variables are missing.
            ValueError: If the symbolic expression is missing.
        """
        if not self._var_names:
            raise ValueError("No variables found in the expression.")
        if not self._aliases:
            raise ValueError("No Python aliases found for variables.")
        if not self._sym_func:
            raise ValueError("No expression has been defined for analysis.")

    def set_coefficient(self, coef: Coefficient) -> None:
        """*set_coefficient()* Configure analysis from a coefficient.

        Args:
            coef (Coefficient): Dimensionless coefficient to analyze.

        Raises:
            ValueError: If the coefficient doesn't have a valid expression.
        """
        if not coef.pi_expr:
            raise ValueError("Coefficient does not have a valid expression.")

        # Set expression
        self._pi_expr = coef.pi_expr
        # parse coefficient expresion
        if coef._pi_expr:
            self._parse_expression(self._pi_expr)

    def _parse_expression(self, expr: str) -> None:
        """*_parse_expression()* Parse the LaTeX expression into a sympy function.

        Args:
            expr (str): LaTeX expression to parse.

        Raises:
            ValueError: If the expression cannot be parsed.
            TypeError: If parsed result is not a valid SymPy expression.
        """
        try:
            # Parse the expression
            parsed_expr = parse_latex(expr)

            # Validate it's actually a SymPy expression
            if parsed_expr is None:
                _msg = "parse_latex returned 'None'."
                raise ValueError(_msg)

            if not isinstance(parsed_expr, sp.Expr):
                _msg = f"Parsed expression is not a SymPy expression: {type(parsed_expr).__name__}."
                raise TypeError(_msg)

            # Create symbol mapping
            maps = create_latex_mapping(expr)
            self._symbols = maps[0]
            self._aliases = maps[1]
            self._latex_to_py = maps[2]
            self._py_to_latex = maps[3]

            # Work with local variable (guaranteed sp.Expr type)
            sym_func: sp.Expr = parsed_expr

            # Substitute LaTeX symbols with Python symbols
            for latex_sym, py_sym in self._symbols.items():
                result = sym_func.subs(latex_sym, py_sym)
                # Ensure result is still an Expr
                if isinstance(result, sp.Expr):
                    sym_func = result
                else:
                    # Convert back to Expr if needed
                    sym_func = sp.sympify(result)

            # Assign the processed expression to the attribute
            self._sym_func = sym_func

            # Get Python variable names and store in _var_names list
            fsyms = sym_func.free_symbols
            self._var_names = sorted([str(s) for s in fsyms])

            # """
            # # OLD code, first version, keep for reference!!!
            # self.results = {
            #     var: lambdify(self._variables, diff(self._sym_fun, var), "numpy")(
            #         *[vals[v] for v in self.variables]
            #     )
            #     for var in self._variables
            # }
            # """
        except (ValueError, TypeError):
            raise  # Re-raise validation errors
        except Exception as e:
            _msg = f"Failed to parse expression: {str(e)}"
            raise ValueError(_msg) from e

    def analyze_symbolically(self,
                             vals: Dict[str, float]) -> Dict[str, float]:
        """*analyze_symbolically()* Perform symbolic sensitivity analysis.

        Args:
            vals (Dict[str, float]): Dictionary mapping variable names to values.

        Returns:
            Dict[str, float]: Sensitivity results for each variable.
        """
        # Validate analysis readiness
        self._validate_analysis_ready()

        # save variable values for the analysis
        self.var_values = vals

        # Check that all required variables are provided
        var_lt = [str(v) for v in self._latex_to_py]
        missing_vars = set(var_lt) - set(list(vals.keys()))
        if missing_vars:
            _msg = f"Missing values for variables: {missing_vars}"
            raise ValueError(_msg)

        # trying symbolic coefficient sensitivity analysis
        try:
            py_to_latex = self._py_to_latex
            results = dict()
            functions = dict()
            if self._var_names:
                for var in self._var_names:
                    # Create lambdify function using Python symbols
                    expr = diff(self._sym_func, var)
                    aliases = [self._aliases[v] for v in self._var_names]
                    # self._exe_func = lambdify(aliases, expr, "numpy")
                    func = lambdify(aliases, expr, "numpy")
                    functions[py_to_latex[var]] = func

                    # Convert back to LaTeX variables for result keys
                    val_args = [vals[py_to_latex[v]] for v in self._var_names]
                    res = functions[py_to_latex[var]](*val_args)
                    results[py_to_latex[var]] = res

            self._exe_func = functions
            self._results = results
            return self._results

        except Exception as e:
            coef = f"{self._sym} = ({self._pi_expr})"
            _msg = f"Error calculating sensitivity for {coef}: {str(e)}"
            raise ValueError(_msg)

    def analyze_numerically(self,
                            vals: List[str],
                            bounds: List[List[float]],
                            iters: int = -1) -> Dict[str, Any]:
        """*analyze_numerically()* Perform numerical sensitivity analysis.

        Args:
            vals (List[str]): List of variable names to analyze.
            bounds (List[List[float]]): Bounds for each variable [min, max].
            iters (int, optional): Number of iterations to use. Defaults to -1.

        Returns:
            Dict[str, Any]: Detailed sensitivity analysis results.
        """
        # Validate analysis readiness
        self._validate_analysis_ready()

        # trying numeric coefficient sensitivity analysis
        try:
            # Validate bounds length matches number of variables
            if len(bounds) != len(self._var_names):
                _msg = f"Expected {len(self._var_names)} "
                _msg += f"bounds (one per variable), got {len(bounds)}"
                raise ValueError(_msg)

            # Set number of samples
            self.experiments = iters
            # Store bounds
            self.var_bounds = bounds

            results: Dict[str, Any] = dict()

            if self._var_names:
                # Set up problem definition for SALib
                problem = {
                    "num_vars": len(vals),
                    "names": self._var_names,
                    "bounds": bounds,
                }

                # Generate samples (domain)
                self._var_domains = sample(problem, self._experiments)
                _len = len(self._var_names)
                self._var_domains = self._var_domains.reshape(-1, _len)

                # Create lambdify function using Python symbols
                aliases = [self._aliases[v] for v in self._var_names]
                func = lambdify(aliases, self._sym_func, "numpy")

                # Store as single callable (overwrite any previous dict)
                self._exe_func = func

                # Type guard: Ensure it's callable before using
                if not callable(self._exe_func):
                    raise TypeError("Failed to create executable function")

                # Evaluate function at sample points
                # using local variable to satisfy type checker
                exe_func = self._exe_func
                # Local variable guaranteed to be Callable
                Y = np.apply_along_axis(lambda v: exe_func(*v),
                                        1,
                                        self._var_domains)
                self._var_ranges = Y.reshape(-1, 1)

                # Perform FAST (Fourier Amplitude Sensitivity Test) analysis
                results = analyze(problem, Y)

                # Convert back to LaTeX variables for result keys
                if "names" in results:
                    py_to_latex = self._py_to_latex
                    results["names"] = [py_to_latex.get(v, v) for v in results["names"]]

            self._results = results
            return self._results

        except Exception as e:
            coef = f"{self._sym} = ({self._pi_expr})"
            _msg = f"Error calculating sensitivity for {coef}: {str(e)}"
            raise ValueError(_msg)

    # Property getters and setters

    @property
    def cat(self) -> str:
        """*cat* Get the analysis category.

        Returns:
            str: Category (SYM, NUM).
        """
        return self._cat

    @cat.setter
    @validate_type(str, allow_none=False)
    @validate_choices(PYDASA_CFG.analitic_modes, case_sensitive=False)
    def cat(self, val: str) -> None:
        """*cat* Set the analysis category.

        Args:
            val (str): Category value.

        Raises:
            ValueError: If category is invalid.
        """
        self._cat = val.upper()

    @property
    def pi_expr(self) -> Optional[str]:
        """*pi_expr* Get the expression to analyze.

        Returns:
            Optional[str]: LaTeX expression.
        """
        return self._pi_expr

    @pi_expr.setter
    @validate_type(str)
    @validate_emptiness()
    @validate_pattern(LATEX_RE, allow_alnum=True)
    def pi_expr(self, val: str) -> None:
        """*pi_expr* Set the expression to analyze.

        Args:
            val (str): LaTeX expression.

        Raises:
            ValueError: If expression is invalid.
        """
        # Update expression
        self._pi_expr = val

        # Parse expression
        self._parse_expression(self._pi_expr)

    @property
    def sym_func(self) -> Optional[sp.Expr]:
        """*sym_func* Get the symbolic expression.

        Returns:
            Optional[sp.Expr]: SymPy symbolic expression.
        """
        return self._sym_func

    @sym_func.setter
    @validate_type(sp.Expr, allow_none=False)
    @validate_custom(lambda self, val: self._validate_sym_expr(val))
    def sym_func(self, val: sp.Expr) -> None:
        """*sym_func* Set the symbolic expression.

        Args:
            val (sp.Expr): SymPy symbolic expression.

        Raises:
            ValueError: If value is not a valid SymPy expression.
        """
        self._sym_func = val

    @property
    def exe_func(self) -> Optional[Union[Callable, Dict[str, Callable]]]:
        """*exe_func* Get the executable function.

        Returns:
            Optional[Union[Callable, Dict[str, Callable]]]: Executable function for numerical evaluation.
        """
        return self._exe_func

    @exe_func.setter
    @validate_custom(lambda self, val: self._validate_exec_func(val))
    def exe_func(self, val: Union[Callable, Dict[str, Callable]]) -> None:
        """*exe_func* Set the executable function(s).

        Args:
            val (Union[Callable, Dict[str, Callable]]): Executable function or
                dictionary of functions.

        Raises:
            ValueError: If value is not a callable or dict of callables.
        """
        self._exe_func = val

    @property
    def variables(self) -> Dict[str, Variable]:
        """*variables* Get the variable symbols dictionary.

        Returns:
            Dict[str, Variable]: Variables symbols dictionary.
        """
        return self._variables

    @variables.setter
    @validate_type(dict, allow_none=False)
    @validate_emptiness()
    @validate_dict_types(str, Variable)
    def variables(self, val: Dict[str, Variable]) -> None:
        """*variables* Set the variable symbols dictionary.

        Args:
            val (Dict[str, Variable]): Variables symbols dictionary.

        Raises:
            ValueError: If variables dictionary is invalid.
        """
        self._variables = val

    @property
    def schema(self) -> Schema:
        """*schema* Get the dimensional schema.

        Returns:
            Schema: Current dimensional schema.
        """
        return self._schema

    @schema.setter
    @validate_type(Schema, allow_none=False)
    def schema(self, val: Schema) -> None:
        """*schema* Set the dimensional schema.

        Args:
            val (Schema): New dimensional schema.

        Raises:
            ValueError: If input is not a Schema instance.
        """
        # Update framework and global configuration
        self._schema = val

    @property
    def symbols(self) -> Dict[sp.Symbol, sp.Symbol]:
        """*symbols* Get the Python symbols for the variables.

        Returns:
            Dict[sp.Symbol, sp.Symbol]: Dictionary mapping variable names to sympy symbols.
        """
        return self._symbols

    @property
    def aliases(self) -> Dict[str, sp.Symbol]:
        """*aliases* Get the Python aliases for the variables.

        Returns:
            Dict[str, sp.Symbol]: Python-compatible variable names.
        """
        return self._aliases

    @property
    def var_names(self) -> List[str]:
        """*var_names* Get the list of variable names extracted from the expression.

        Returns:
            List[str]: Ordered list of variable names.
        """
        return self._var_names.copy()

    @property
    def coefficient(self) -> Optional[Coefficient]:
        """*coefficient* Get the dimensionless coefficient specification.

        Returns:
            Optional[Coefficient]: Coefficient specification.
        """
        return self._coefficient

    @coefficient.setter
    @validate_type(Coefficient, allow_none=True)
    def coefficient(self, val: Optional[Coefficient]) -> None:
        """*coefficient* Set the dimensionless coefficient specification.

        Args:
            val (Optional[Coefficient]): Coefficient specification.

        Raises:
            ValueError: If value is not a Coefficient instance.
        """
        self._coefficient = val

    @property
    def latex_to_py(self) -> Dict[str, str]:
        """*latex_to_py* Get the mapping from LaTeX symbols to Python compatible names.

        Returns:
            Dict[str, str]: LaTeX to Python symbol mapping.
        """
        return self._latex_to_py.copy()

    @property
    def py_to_latex(self) -> Dict[str, str]:
        """*py_to_latex* Get the mapping from Python compatible names to LaTeX symbols.

        Returns:
            Dict[str, str]: Python to LaTeX symbol mapping.
        """
        return self._py_to_latex.copy()

    @property
    def var_bounds(self) -> List[List[float]]:
        """*var_bounds* Get the min/max bounds for each variable.

        Returns:
            List[List[float]]: Bounds for each variable.
        """
        return self._var_bounds.copy()

    @var_bounds.setter
    @validate_type(list, allow_none=False)
    @validate_emptiness()
    def var_bounds(self, val: List[List[float]]) -> None:
        """*var_bounds* Set the min/max bounds for each variable.

        Args:
            val (List[List[float]]): Bounds for each variable [min, max].

        Raises:
            ValueError: If bounds list is invalid or empty.
        """
        self._var_bounds = val

    @property
    def var_values(self) -> Dict[str, float]:
        """*var_values* Get the values for symbolic analysis.

        Returns:
            Dict[str, float]: Variable values for symbolic analysis.
        """
        return self._var_values.copy()

    @var_values.setter
    @validate_type(dict, allow_none=False)
    @validate_emptiness()
    def var_values(self, val: Dict[str, float]) -> None:
        """*var_values* Set the values for symbolic analysis.

        Args:
            val (Dict[str, float]): Variable values for symbolic analysis.

        Raises:
            ValueError: If values dictionary is invalid or empty.
        """
        self._var_values = val

    @property
    def var_domains(self) -> Optional[np.ndarray]:
        """*var_domains* Get the sample domain for numerical analysis.

        Returns:
            Optional[np.ndarray]: Sample domain (inputs) or None if not computed.
        """
        if self._var_domains is None:
            return None
        return self._var_domains.copy()

    @property
    def var_ranges(self) -> Optional[np.ndarray]:
        """*var_ranges* Get the sample value range for numerical analysis.

        Returns:
            Optional[np.ndarray]: Sample value range (results) or None if not computed.
        """
        if self._var_ranges is None:
            return None
        return self._var_ranges.copy()

    @property
    def results(self) -> Dict[str, Any]:
        """*results* Get the analysis results.

        Returns:
            Dict[str, Any]: Analysis results dictionary.
        """
        return self._results.copy()

    @property
    def experiments(self) -> int:
        """*experiments* Get the number of experiments for analysis.

        Returns:
            int: Number of experiments.
        """
        return self._experiments

    @experiments.setter
    @validate_type(int, allow_none=False)
    @validate_index(allow_zero=False)
    def experiments(self, val: int) -> None:
        """*experiments* Set the number of experiments for analysis.

        Args:
            val (int): Number of experiments (must be >= 1).

        Raises:
            ValueError: If experiments value is invalid or less than 1.
        """
        self._experiments = val

    def clear(self) -> None:
        """*clear()* Reset all attributes to default values.

        Resets all sensitivity analysis properties to their initial state.
        """
        # Reset parent class attributes (Foundation)
        super().clear()
        self._sym = f"SENS_{{\\Pi_{{{self._idx}}}}}"

        # Reset sensitivity-specific attributes
        self._cat = AnaliticMode.SYM.value
        self._pi_expr = None
        self._sym_func = None
        self._exe_func = None
        self._variables = {}
        self._var_names = []
        self._schema = Schema()
        self._coefficient = None
        self._symbols = {}
        self._aliases = {}
        self._latex_to_py = {}
        self._py_to_latex = {}
        self._var_bounds = []
        self._var_values = {}
        self._var_domains = None
        self._var_ranges = None
        self._experiments = -1
        self._results = {}

    def to_dict(self) -> Dict[str, Any]:
        """*to_dict()* Convert sensitivity analysis to dictionary representation.

        Returns:
            Dict[str, Any]: Dictionary representation of sensitivity analysis.
        """
        result = {}

        # Get all dataclass fields
        for f in fields(self):
            attr_name = f.name
            attr_value = getattr(self, attr_name)

            # Skip numpy arrays (not JSON serializable without special handling)
            if isinstance(attr_value, np.ndarray):
                # Convert to list for JSON compatibility
                attr_value = attr_value.tolist()

            # Skip SymPy objects (not JSON serializable)
            if isinstance(attr_value, sp.Expr):
                # Convert to string representation
                attr_value = str(attr_value)
            elif isinstance(attr_value, dict):
                # Handle dictionaries with SymPy keys/values
                cleaned_dict = {}
                for k, v in attr_value.items():
                    # Convert SymPy keys to strings
                    key = str(k) if isinstance(k, sp.Basic) else k
                    # Convert SymPy values to strings
                    value = str(v) if isinstance(v, sp.Basic) else v
                    cleaned_dict[key] = value
                attr_value = cleaned_dict

            # Skip callables (can't be serialized)
            if callable(attr_value) and not isinstance(attr_value, type):
                continue

            # Skip Schema objects (complex nested structure)
            if attr_name == "_schema":
                continue

            # Remove leading underscore from private attributes
            if attr_name.startswith("_"):
                clean_name = attr_name[1:]  # Remove first character
            else:
                clean_name = attr_name

            result[clean_name] = attr_value

        return result

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Sensitivity:
        """*from_dict()* Create sensitivity analysis from dictionary representation.

        Args:
            data (Dict[str, Any]): Dictionary representation of sensitivity analysis.

        Returns:
            Sensitivity: New sensitivity analysis instance.
        """
        # Get all valid field names from the dataclass
        field_names = {f.name for f in fields(cls)}

        # Map keys without underscores to keys with underscores
        mapped_data = {}

        for key, value in data.items():
            # Try the key as-is first (handles both _idx and name)
            if key in field_names:
                mapped_data[key] = value
            # Try adding underscore prefix (handles idx -> _idx)
            elif f"_{key}" in field_names:
                mapped_data[f"_{key}"] = value
            # Try removing underscore prefix (handles _name -> name if needed)
            elif key.startswith("_") and key[1:] in field_names:
                mapped_data[key[1:]] = value
            else:
                # Use as-is for unknown keys (will be validated by dataclass)
                mapped_data[key] = value

        # Convert lists back to numpy arrays for domain/range attributes
        for array_key in ["var_domains", "_var_domains", "var_ranges", "_var_ranges"]:
            if array_key in mapped_data and isinstance(mapped_data[array_key], list):
                mapped_data[array_key] = np.array(mapped_data[array_key])

        return cls(**mapped_data)



# -*- coding: utf-8 -*-
"""
Module practical.py
===========================================

Module for **MonteCarloSimulation** to manage the Monte Carlo experiments in *PyDASA*.

This module provides classes for managing Monte Carlo simulations of the dimensionless coefficients.

Classes:
    **MonteCarloSimulation**: Manages Monte Carlo simulation runs, including configuration and execution of the experiments.

*IMPORTANT:* Based on the theory from:

    # H.Gorter, *Dimensionalanalyse: Eine Theoririe der physikalischen Dimensionen mit Anwendungen*
"""
# Standard library imports
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict, List, Any, Optional, cast

# import random
# import re, Optional, Callable

# Third-party imports
import numpy as np
from numpy.typing import NDArray

# Import validation base classes
from pydasa.core.basic import Foundation
from pydasa.workflows.basic import WorkflowBase

# Import related classes
from pydasa.dimensional.buckingham import Coefficient
from pydasa.analysis.simulation import MonteCarlo

# Import utils
from pydasa.serialization.parser import latex_to_python

# Import validation decorators
from pydasa.validations.decorators import validate_type
from pydasa.validations.decorators import validate_range
from pydasa.validations.decorators import validate_choices

# Import global configuration
# from pydasa.core.parameter import Variable
# from pydasa.core.setup import Frameworks
from pydasa.core.setup import SimulationMode
# Import configuration
from pydasa.core.setup import PYDASA_CFG


@dataclass
class MonteCarloSimulation(Foundation, WorkflowBase):
    """**MonteCarloSimulation** class for managing Monte Carlo simulations in *PyDASA*.

    Manages the creation, configuration, and execution of Monte Carlo simulations of dimensionless coefficients.

    Args:
        Foundation (Foundation): Inherits common validation logic.
        WorkflowBase (WorkflowBase): Inherits workflow basic functionalities.

    Attributes:
        # From Foundation (Identification and Classification):
            _name (str): User-friendly name of the Monte Carlo simulation handler.
            description (str): Brief summary of the Monte Carlo simulation handler.
            _idx (int): Index/precedence of the simulation handler.
            _sym (str): Symbol representation (LaTeX or alphanumeric).
            _alias (str): Python-compatible alias for use in code.
            _fwk (str): Frameworks context (PHYSICAL, COMPUTATION, SOFTWARE, CUSTOM).

        # From WorkflowBase (Common Workflow Components):
            _variables (Dict[str, Variable]): All available parameters/variables in the model (*Variable*). Accepts Variable instances or dicts.
            _schema (Optional[Schema]): Dimensional framework schema for the workflow. After __post_init__, this will always be a Schema instance (inherited but not actively used in MonteCarloSimulation).
            _coefficients (Dict[str, Coefficient]): All available coefficients in the model (*Coefficient*). Accepts Coefficient instances or dicts.
            _results (Dict[str, Dict[str, Any]]): Consolidated results from Monte Carlo simulations.
            _is_solved (bool): Flag indicating if the simulation workflow has been completed.

        # Specific to MonteCarloSimulation:
            _cat (str): Category of simulation run (DIST, DATA).
            _distributions (Dict[str, Dict[str, Any]]): All distribution functions used in the simulations (specific name, parameters, and function).
            _experiments (int): Number of simulations to run. Default is -1.
            _simulations (Dict[str, MonteCarlo]): All Monte Carlo simulations performed for each coefficient.
            _shared_cache (Dict[str, NDArray[np.float64]]): In-memory cache for simulation data shared between coefficients.
    """

    # ========================================================================
    # MonteCarloSimulation Specific Attributes
    # ========================================================================

    # :attr: _cat
    _cat: str = SimulationMode.DIST.value
    """Category of simulation run (DIST, DATA)."""

    # Simulation configuration
    # :attr: _distributions
    _distributions: Dict[str, Dict[str, Any]] = field(default_factory=dict)
    """Variable sampling distributions and specifications for simulations (specific name, parameters, and function)."""

    # :attr: _experiments
    _experiments: int = -1
    """Number of simulation to run."""

    # Simulation Management
    # :attr: _shared_cache
    _shared_cache: Dict[str, NDArray[np.float64]] = field(default_factory=dict)
    """In-memory cache for simulation data between coefficients."""

    # :attr: _simulations
    _simulations: Dict[str, MonteCarlo] = field(default_factory=dict)
    """Dictionary of Monte Carlo simulations."""

    def __post_init__(self) -> None:
        """*__post_init__()* Initializes the Monte Carlo handler."""
        # Initialize from base class
        super().__post_init__()

        # Set default symbol if not specified
        if not self._sym:
            self._sym = f"MCS_{{\\Pi_{{{self._idx}}}}}" if self._idx >= 0 else "MCS_\\Pi_{-1}"

        if not self._alias:
            self._alias = latex_to_python(self._sym)

        # Set name and description if not already set
        if not self.name:
            self.name = f"Monte Carlo Simulation Handler {self._idx}"

        if not self.description:
            coef_keys = ", ".join(self._coefficients.keys()) if self._coefficients else "no coefficients"
            self.description = f"Manages Monte Carlo simulations for [{coef_keys}] coefficients."

        # Ensure mem_cache is always initialized
        if self._shared_cache is None:
            self._shared_cache = {}

    def configure_distributions(self) -> None:
        """*configure_distributions()* Configures sampling distributions for all variables. Sets up distribution specifications (type, parameters, functions) for each variable that will be used during Monte Carlo simulation.

        Raises:
            ValueError: If variables are not defined.
            ValueError: If distribution specifications are invalid.
        """
        if len(self._distributions) == 0:
            self._config_distributions()

    def configure_simulations(self) -> None:
        """*configure_simulations()* Configures Monte Carlo simulation objects for each coefficient. Creates MonteCarlo instances for each coefficient with appropriate distributions and dependencies. Requires distributions to be configured first.

        Raises:
            ValueError: If distributions are not configured.
            ValueError: If coefficients or variables are not defined.
        """
        if len(self._simulations) == 0:
            self._config_simulations()

    def create_simulations(self) -> None:
        """*create_simulations()* Configures distributions and simulations if not already set. This is a convenience method that orchestrates the full configuration process. It calls configure_distributions() and configure_simulations() in sequence.

        Raises:
            ValueError: If variables or coefficients are not defined.
            ValueError: If distribution specifications are invalid.
        """
        self.configure_distributions()
        self.configure_simulations()

    # ========================================================================
    # Helper Methods
    # ========================================================================

    def _validate_coefficient_vars(self,
                                   coef: Coefficient,
                                   pi_sym: str) -> Dict[str, Any]:
        """*_validate_coefficient_vars()* Validates and returns coefficient's var_dims.

        Args:
            coef (Coefficient): The coefficient to validate.
            pi_sym (str): The coefficient symbol for error messages.

        Returns:
            Dict[str, Any]: The validated var_dims dictionary.

        Raises:
            ValueError: If var_dims is None or missing.
        """
        if not hasattr(coef, 'var_dims'):
            _msg = f"Coefficient '{pi_sym}' missing var_dims attribute."
            raise ValueError(_msg)

        var_dims = coef.var_dims
        if var_dims is None:
            _msg = f"Coefficient '{pi_sym}' has None var_dims. "
            _msg += "Ensure the coefficient was properly initialized."
            raise ValueError(_msg)

        if not isinstance(var_dims, dict):
            _msg = f"Coefficient '{pi_sym}' var_dims must be a dictionary. "
            _msg += f"Got: {type(var_dims).__name__}"
            raise TypeError(_msg)

        return var_dims

    # ========================================================================
    # Configuration Methods
    # ========================================================================

    def _config_distributions(self) -> None:
        """*_config_distributions()* Creates the Monte Carlo distributions for each variable.

        Raises:
            ValueError: If the distribution specifications are invalid.
        """
        # Clear existing distributions
        self._distributions.clear()

        # Validate variables exist before processing
        if not self._variables:
            _msg = "Cannot configure distributions: no variables defined."
            raise ValueError(_msg)

        for var in self._variables.values():
            sym = var.sym

            # Skip if already configured
            if sym in self._distributions:
                continue

            # Collect specs for better error reporting
            specs = {
                "dist_type": var.dist_type,
                "dist_params": var.dist_params,
                "dist_func": var.dist_func
            }

            # Validate distribution specifications
            if not any(specs.values()):
                _msg = f"Invalid distribution for variable '{sym}'. "
                _msg += f"Incomplete specifications provided: {specs}"
                raise ValueError(_msg)

            # Store distribution configuration
            self._distributions[sym] = {
                "depends": var.depends,
                "dtype": var.dist_type,
                "params": var.dist_params,
                "func": var.dist_func
            }

    def _get_distributions(self,
                           var_keys: List[str]) -> Dict[str, Dict[str, Any]]:
        """*_get_distributions()* Retrieves the distribution specifications for a list of variable keys.

        Args:
            var_keys (List[str]): List of variable keys.

        Returns:
            Dict[str, Dict[str, Any]]: Dictionary of distribution specifications.

        Raises:
            ValueError: If required distributions are missing.
        """
        # Filter distributions for requested variables
        dist = {k: v for k, v in self._distributions.items() if k in var_keys}

        # Warn about missing distributions
        missing = [k for k in var_keys if k not in dist]

        if missing:
            _msg = f"Missing distributions for variables: {missing}. "
            _msg += "Ensure _config_distributions() has been called."
            raise ValueError(_msg)

        return dist

    def _get_dependencies(self, var_keys: List[str]) -> Dict[str, List[str]]:
        """*_get_dependencies()* Retrieves variable dependencies for a list of variable keys.

        Args:
            var_keys (List[str]): List of variable keys.

        Returns:
            Dict[str, List[str]]: Dictionary mapping variable symbols to their dependencies.
        """
        deps = {
            k: v.depends for k, v in self._variables.items() if k in var_keys
        }
        return deps

    def _init_shared_cache(self) -> None:
        """*_init_shared_cache()* Initialize shared cache for all variables."""
        # Only initialize if experiments is positive
        if self._experiments < 0:
            return

        # Initialize cache for each variable once
        for var_sym in self._variables.keys():
            self._shared_cache[var_sym] = np.full((self._experiments, 1),
                                                  np.nan,
                                                  dtype=np.float64)

    def _config_simulations(self) -> None:
        """*_config_simulations()* Sets up Monte Carlo simulation objects for each coefficient to be analyzed.

        Creates a MonteCarlo instance for each coefficient with appropriate distributions and dependencies.

        Raises:
            ValueError: If coefficients or variables are not properly configured.
        """
        # Validate prerequisites
        if not self._coefficients:
            _msg = "Cannot configure simulations: no coefficients defined."
            raise ValueError(_msg)

        if not self._variables:
            _msg = "Cannot configure simulations: no variables defined."
            raise ValueError(_msg)

        if not self._distributions:
            _msg = "Cannot configure simulations: distributions not defined. "
            raise ValueError(_msg)

        # Clear existing simulations
        self._simulations.clear()

        # Initialize shared cache once
        if not self._shared_cache:
            self._init_shared_cache()

        # Create simulations for each coefficient
        for i, (pi, coef) in enumerate(self._coefficients.items()):
            # Validate coefficient before processing
            var_dims = self._validate_coefficient_vars(coef, pi)

            # Extract variables from the coefficient's expression
            vars_in_coef = list(var_dims.keys())

            # Skip coefficients with no variables
            if not vars_in_coef:
                _msg = f"Coefficient '{pi}' has no variables in expression. Skipping simulation."
                print(f"Warning: {_msg}")
                continue

            try:
                # Create Monte Carlo simulation
                sim = MonteCarlo(
                    _idx=i,
                    _sym=f"MC_{{{coef.sym}}}",
                    _fwk=self._fwk,
                    _cat=self._cat,
                    _pi_expr=coef.pi_expr,
                    _coefficient=coef,
                    _variables=self._variables,
                    # _simul_cache=self._shared_cache,
                    _experiments=self._experiments,
                    _name=f"Monte Carlo Simulation for {coef.name}",
                    description=f"Monte Carlo simulation for {coef.sym}",
                )

                # Configure with coefficient
                sim.set_coefficient(coef)

                # Get distributions with validation
                sim._distributions = self._get_distributions(vars_in_coef)
                sim._dependencies = self._get_dependencies(vars_in_coef)

                # CRITICAL: Share the cache reference
                sim._simul_cache = self._shared_cache

                # Add to simulations dictionary
                self._simulations[pi] = sim

            except Exception as e:
                _msg = f"Failed to create simulation for '{pi}': {str(e)}"
                raise RuntimeError(_msg) from e

    # ========================================================================
    # Simulation Execution Methods
    # ========================================================================

    def run(self, iters: Optional[int] = None) -> None:
        """*run()* Runs the Monte Carlo simulations.

        Args:
            iters (Optional[int]): Number of iterations (experiments) to run.
                If None, uses self._experiments value. Defaults to None.

        Raises:
            ValueError: If simulations are not configured.
            ValueError: If a required simulation is not found.
        """
        # Validate simulations exist
        if not self._simulations:
            _msg = "No simulations configured. Call create_simulations() first."
            raise ValueError(_msg)

        # Use default if not specified
        if iters is not None:
            self._experiments = iters

        #  Validate experiments
        if self._experiments < 1:
            _msg = f"Experiments must be positive. Got: {iters}"
            raise ValueError(_msg)

        # Initialize shared cache BEFORE running simulations
        if not self._shared_cache:
            for var_sym in self._variables.keys():
                self._shared_cache[var_sym] = np.full((self._experiments, 1),
                                                      np.nan,
                                                      dtype=np.float64)

        # print("----------")
        # print(f"_shared_cache keys: {self._shared_cache.keys()}")
        # # Assign shared cache to ALL simulations
        # for sim in self._simulations.values():
        #     sim._simul_cache = self._shared_cache

        results = {}

        for sym in self._coefficients:
            # Get the simulation object
            sim = self._simulations.get(sym)
            if not sim:
                _msg = f"Simulation for coefficient '{sym}' not found. "
                _msg += "Ensure _config_simulations() completed successfully."
                raise ValueError(_msg)

            try:
                # print("-----------------------------------")
                # print(f"_shared_cache status:\n {self._shared_cache}")
                # print("-----------------------------------")

                # ✅ Use shared cache
                sim._simul_cache = self._shared_cache

                # Run the simulation
                sim.run(self._experiments)

                # Store comprehensive results
                res = {
                    "inputs": sim.data,
                    "results": sim.results,
                    "statistics": sim.statistics,
                }

                # Store results
                results[sym] = res

            except Exception as e:
                _msg = f"Simulation failed for coefficient '{sym}': {str(e)}"
                raise RuntimeError(_msg) from e

        self._results = results

    def run_simulation(self, iters: Optional[int] = None) -> None:
        """*run_simulation()* Convenience method to configure and run Monte Carlo simulations.

        This method combines create_simulations() and run() into a single call. It automatically sets up distributions and simulations if needed, then runs the simulation.

        Args:
            iters (Optional[int]): Number of iterations (experiments) to run.
                If None, uses self._experiments value. Defaults to None.

        Example:
            >>> mc_handler = MonteCarloSimulation(...)
            >>> mc_handler.run_simulation(iters=10000)
        """
        # Create simulations if not already configured
        self.create_simulations()

        # Run the simulations
        self.run(iters=iters)

    # ========================================================================
    # Getter Methods
    # ========================================================================

    def get_simulation(self, name: str) -> MonteCarlo:
        """*get_simulation()* Get a simulation by name.

        Args:
            name (str): Name of the simulation.

        Returns:
            MonteCarlo: The requested simulation.

        Raises:
            ValueError: If the simulation doesn't exist.
        """
        if name not in self._simulations:
            available = ", ".join(self._simulations.keys())
            _msg = f"Simulation '{name}' does not exist. "
            _msg += f"Available: {available}"
            raise ValueError(_msg)

        return self._simulations[name]

    def get_distribution(self, name: str) -> Dict[str, Any]:
        """*get_distribution()* Get the distribution by name.

        Args:
            name (str): Name of the distribution.

        Returns:
            Dict[str, Any]: The requested distribution.

        Raises:
            ValueError: If the distribution doesn't exist.
        """
        if name not in self._distributions:
            available = ", ".join(self._distributions.keys())
            _msg = f"Distribution '{name}' does not exist. "
            _msg += f"Available: {available}"
            raise ValueError(_msg)

        return self._distributions[name]

    def get_results(self, name: str) -> Dict[str, Any]:
        """*get_results()* Get the results of a simulation by name.

        Args:
            name (str): Name of the simulation.

        Returns:
            Dict[str, Any]: The results of the requested simulation.

        Raises:
            ValueError: If the results for the simulation don't exist.
        """
        if name not in self._results:
            available = ", ".join(self._results.keys())
            _msg = f"Results for simulation '{name}' do not exist. "
            _msg += f"Available: {available}"
            raise ValueError(_msg)

        return self._results[name]

    # ========================================================================
    # Property Getters and Setters
    # ========================================================================

    @property
    def cat(self) -> str:
        """*cat* Get the Monte Carlo Simulation category.

        Returns:
            str: Category (DIST, DATA).
        """
        return self._cat

    @cat.setter
    @validate_type(str)
    @validate_choices(PYDASA_CFG.simulation_modes)
    def cat(self, val: str) -> None:
        """*cat* Set the Monte Carlo Simulation category.

        Args:
            val (str): Category value.

        Raises:
            ValueError: If category is invalid.
        """
        self._cat = val.upper()

    @property
    def experiments(self) -> int:
        """*experiments* Get the number of experiments.

        Returns:
            int: Number of experiments to run.
        """
        return self._experiments

    @experiments.setter
    @validate_type(int)
    @validate_range(min_value=1)
    def experiments(self, val: int) -> None:
        """*experiments* Set the number of experiments.

        Args:
            val (int): Number of experiments.

        Raises:
            ValueError: If value is not positive.
        """
        self._experiments = val

    @property
    def simulations(self) -> Dict[str, MonteCarlo]:
        """*simulations* Get the dictionary of Monte Carlo simulations.

        Returns:
            Dict[str, MonteCarlo]: Dictionary of Monte Carlo simulations.
        """
        return self._simulations.copy()

    @property
    def is_configured(self) -> bool:
        """*is_configured* Check if simulations are configured and ready to run.

        Returns:
            bool: True if both distributions and simulations are configured.
        """
        return len(self._distributions) > 0 and len(self._simulations) > 0

    @property
    def has_results(self) -> bool:
        """*has_results* Check if simulation results are available.

        Returns:
            bool: True if simulations have been run and results exist.
        """
        return len(self._results) > 0

    # ========================================================================
    # Utility Methods
    # ========================================================================

    def reset(self) -> None:
        """*reset()* Reset simulation state while preserving input configuration.

        Clears results, solved state, and simulation-specific attributes (simulations, distributions, cache) while keeping variables, schema, and coefficients.
        """
        # Clear MonteCarloSimulation specific attributes
        self._simulations.clear()
        self._distributions.clear()
        self._shared_cache.clear()

        # Clear only results and is_solved, preserve coefficients (they're input config)
        self._results.clear()
        self._is_solved = False

    def clear(self) -> None:
        """*clear()* Clear all attributes to default values.

        Resets all simulation properties to their initial state, including variables, coefficients, and results from WorkflowBase.
        """
        # Reset MonteCarloSimulation specific attributes first
        self._cat = SimulationMode.DIST.value
        self._simulations = {}
        self._distributions = {}
        self._shared_cache = {}
        self._experiments = -1

        # Call both parent classes' clear methods explicitly for multiple inheritance
        Foundation.clear(self)
        WorkflowBase.clear(self)

        # Reset symbol after parent clears
        self._sym = f"MCS_{{\\Pi_{{{self._idx}}}}}"

    def to_dict(self) -> Dict[str, Any]:
        """*to_dict()* Convert the handler's state to a dictionary.

        Returns:
            Dict[str, Any]: Dictionary representation of the handler's state.
        """
        # Get base serialization from WorkflowBase
        result = super().to_dict()

        # Add MonteCarloSimulation-specific attributes
        result["cat"] = self._cat
        result["experiments"] = self._experiments
        result["simulations"] = {
            k: v.to_dict() for k, v in self._simulations.items()
        }
        result["distributions"] = self._distributions

        return result

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> MonteCarloSimulation:
        """*from_dict()* Create a MonteCarloSimulation instance from a dictionary.

        Args:
            data (Dict[str, Any]): Dictionary containing the handler's state.

        Returns:
            MonteCarloSimulation: New instance of MonteCarloSimulation.
        """
        # Use parent class to handle base deserialization
        instance = cast(MonteCarloSimulation, super().from_dict(data))

        # Handle MonteCarloSimulation-specific attributes
        instance._cat = data.get("cat", SimulationMode.DIST.value)
        instance._experiments = data.get("experiments", -1)
        instance._distributions = data.get("distributions", {})

        # Reconstruct MonteCarlo simulations if present
        simulations_data = data.get("simulations", {})
        if simulations_data:
            instance._simulations = {
                k: MonteCarlo.from_dict(v) for k, v in simulations_data.items()
            }

        return instance



# -*- coding: utf-8 -*-
"""
Module constants.py
===========================================

This module specifies the Fundamental Dimensional Units (FDUs) available by default in *PyDASA*. The default framework are the Physical FDUs.

The three main types of FDUs are:
    1. Physical FDUs: in `PHY_FDU_PREC_DT` representing the standard physical dimensions.
    2. Digital FDUs: in `COMPU_FDU_PREC_DT` representing dimensions relevant to computation.
    3. Software Architecture FDUs: in `SOFT_FDU_PREC_DT` representing dimensions specific to software architecture.

The fourth FDU framework is:
    4. Custom FDUs: user-defined FDUs that can be specified as needed.
"""
# python native modules

# global variables

# Default config folder name + settings
# :attr: DFLT_CFG_FOLDER
DFLT_CFG_FOLDER: str = "cfg"
# :attr: DFLT_CFG_FILE
DFLT_CFG_FILE: str = "default.json"
"""
*PyDASA* default configuration folder and file names.
"""



# -*- coding: utf-8 -*-
"""
Module latex.py
===========================================

Module for default global variables and comparison functions for use by all *PyDASA* and its Data Structs.
"""
# python native modules
# from dataclasses import dataclass
# from typing import TypeVar
from typing import Tuple, Dict, List, Optional, Any, Callable, Sequence
import re

# custom modules
from sympy.parsing.latex import parse_latex
from sympy import Symbol, symbols

# import global variables
from pydasa.validations.patterns import LATEX_RE
from pydasa.validations.patterns import PI_POW_RE
from pydasa.validations.patterns import PI_COEF_RE
from pydasa.validations.patterns import BASIC_OPS_RE
# Global vars for special Latex symbos and functions to ignore
from pydasa.validations.patterns import IGNORE_EXPR


# Latex Parsing Functions

def latex_to_python(expr: str) -> str:
    """*latex_to_python()* Convert a LaTeX expression to a Python-compatible string.

    Args:
        expr (str): The LaTeX expression to convert.

    Returns:
        str: The Python-compatible string.
    """
    # Replace LaTeX subscript with Python style
    if expr.isalnum():
        return expr
    # TODO this regex doesnt work, check latter
    # ans = re.sub(r"\\([a-zA-Z]+)_{(\d+)}", r"\1_\2", expr)
    alias = expr.replace("\\", "")
    alias = alias.replace("_{", "_").replace("}", "")
    return alias


def extract_latex_vars(expr: str) -> Tuple[Dict[str, str], Dict[str, str]]:
    """*extract_latex_vars()* Extract variable names in LaTeX format with their Python equivalents.

    Args:
        expr (str): The LaTeX expression to parse.

    Returns:
        tuple [dict]: A tuple containing two dictionaries:
                - The first dictionary maps LaTeX variable names to their Python equivalents.
                - The second dictionary maps Python variable names to their LaTeX equivalents.
    """
    # Extract latex variable names with regex
    matches = re.findall(LATEX_RE, expr)

    # Filter out ignored LaTeX commands
    matches = [m for m in matches if m not in IGNORE_EXPR]

    # Create mappings both ways
    latex_to_py = {}
    py_to_latex = {}

    for m in matches:
        # Keep original LaTeX notation for external reference
        latex_var = m
        # Convert to Python style for internal use
        py_var = m.lstrip("\\")
        py_var = py_var.replace("_{", "_")
        py_var = py_var.replace("}", "")

        latex_to_py[latex_var] = py_var
        py_to_latex[py_var] = latex_var

    return latex_to_py, py_to_latex


def create_latex_mapping(expr: str) -> Tuple[Dict[Symbol, Symbol],  # symbol_map
                                             Dict[str, Symbol],  # py_symbol_map
                                             Dict[str, str],     # latex_to_py
                                             Dict[str, str]      # py_to_latex
                                             ]:
    """*create_latex_mapping()* Create a mapping between LaTeX symbols and Python symbols.

    Args:
        expr (str): The LaTeX expression to parse.

    Returns:
        tuple[dict]: A tuple containing:
            - A dictionary mapping LaTeX symbols to Python symbols for internal substitution.
            - A dictionary mapping Python variable names to their corresponding sympy symbols for lambdify.
            - A dictionary mapping LaTeX variable names to their Python equivalents.
            - A dictionary mapping Python variable names to their LaTeX equivalents.
    """
    # Get LaTeX<->Python variable mappings
    latex_to_py, py_to_latex = extract_latex_vars(expr)

    # Parse to get LaTeX symbols
    sym_expr = parse_latex(expr)

    # Create mapping for sympy substitution
    symbol_map = {}         # For internal substitution
    py_symbol_map = {}      # For lambdify

    for latex_sym in sym_expr.free_symbols:
        latex_name = str(latex_sym)

        # Find corresponding Python name
        for latex_var, py_var in latex_to_py.items():
            # Check for various forms of equivalence
            con1 = (latex_name == latex_var)
            con2 = (latex_name == py_var)
            con3 = (latex_name.replace("_{", "_").replace("}", "") == py_var)
            if con1 or con2 or con3:
                # Create symbol for this variable
                sym = symbols(py_var)
                # Store mappings
                symbol_map[latex_sym] = sym  # For substitution
                py_symbol_map[py_var] = sym  # For lambdify args
                break

    return symbol_map, py_symbol_map, latex_to_py, py_to_latex


# ============================================================================
# Dimensional Expression Parsing Functions
# TODO improve using sympy + pint later
# ============================================================================

def extract_coeff_syms(expr: str, patt: str = PI_COEF_RE) -> List[str]:
    """*extract_coeff_syms()* Extract Pi coefficient symbols from expression.

    Args:
        expr (str): Mathematical expression containing Pi coefficients.
        patt (str): Regex pattern to match coefficients. Defaults to Pi notation.

    Returns:
        List[str]: List of unique coefficient symbols (e.g., ['\\Pi_{0}', '\\Pi_{1}']).

    Example:
        >>> extract_coeff_syms("\\Pi_{0} * \\Pi_{1}**(-1)")
        ['\\Pi_{0}', '\\Pi_{1}']
    """
    # Use finditer to get full match instead of captured groups
    matches = re.finditer(patt, expr)
    # Remove duplicates while preserving order
    seen = set()
    unique = []
    for match in matches:
        sym = match.group(0)  # Get full match, not captured group
        if sym not in seen:
            seen.add(sym)
            unique.append(sym)
    return unique


def extract_powered_coeffs(expr: str,
                           pow_patt: str = PI_POW_RE) -> Tuple[str, Dict[str, Tuple[str, float]]]:
    """*extract_powered_coeffs()* Parse and replace power expressions with placeholders.

    Identifies patterns like \\Pi_{0}**(-1) and replaces them with temporary
    placeholders for easier parsing.

    Args:
        expr (str): Mathematical expression with potential power operations.
        pow_patt (str): Regex pattern to match power expressions.

    Returns:
        Tuple[str, Dict[str, Tuple[str, float]]]:
            - Modified expression with placeholders
            - Dictionary mapping placeholders to (coefficient_symbol, power) tuples

    Raises:
        ValueError: If power expression is invalid.

    Example:
        >>> expr = "\\Pi_{0}**(-1) * \\Pi_{1}**(2)"
        >>> new_expr, transforms = extract_powered_coeffs(expr)
        >>> # new_expr: "__PHLR_0__ * __PHLR_1__"
        >>> # transforms: {"__PHLR_0__": ("\\Pi_{0}", -1), "__PHLR_1__": ("\\Pi_{1}", 2)}
    """
    pow_transforms = {}
    n_placeholder = 0

    for match in re.finditer(pow_patt, expr):
        coef_sym = match.group(1)
        pow_str = match.group(2)

        try:
            power = float(pow_str)
        except ValueError:
            raise ValueError(f"Invalid power expression: {pow_str}")

        # Create placeholder and store transformation
        placeholder = f"__PHLR_{n_placeholder}__"
        pow_transforms[placeholder] = (coef_sym, power)
        expr = expr.replace(match.group(0), placeholder, 1)
        n_placeholder += 1

    return expr, pow_transforms


def extract_num_consts(expr: str) -> List[float]:
    """*extract_num_consts()* Extract numeric constants from expression.

    Args:
        expr (str): Mathematical expression containing numeric values.

    Returns:
        List[float]: List of numeric constants found in expression.

    Example:
        >>> extract_num_consts("2 * \\Pi_{0} + 0.5")
        [2.0, 0.5]
    """
    pattern = r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?"
    matches = re.findall(pattern, expr)
    return [float(m) for m in matches]


def compute_dims_opn(dim_col_a: Sequence[int | float],
                     dim_col_b: Sequence[int | float],
                     opn: str) -> List[int]:
    """*compute_dims_opn()* Compute result of dimensional operation.

    Applies dimensional analysis rules:
        - Multiplication (*): Add exponents
        - Division (/): Subtract exponents
        - Addition (+) / Subtraction (-): Result is dimensionless (all zeros)

    Args:
        dim_col_a (Sequence[int | float]): First dimensional column.
        dim_col_b (Sequence[int | float]): Second dimensional column.
        opn (str): Operation to perform ('*', '/', '+', '-').

    Returns:
        List[int]: Resulting dimensional column.

    Example:
        >>> compute_dims_opn([1, 2, 0], [0, 1, -1], "*")
        [1, 3, -1]  # L^1 * M^2 * L^0 * M^1 * T^-1 = L^1 * M^3 * T^-1
    """
    if opn == "*":
        return [int(a + b) for a, b in zip(dim_col_a, dim_col_b)]
    elif opn == "/":
        return [int(a - b) for a, b in zip(dim_col_a, dim_col_b)]
    elif opn in ("+", "-"):
        # Addition/subtraction: result is dimensionless
        return [0] * len(dim_col_a)
    else:
        raise ValueError(f"Unknown operation: {opn}")


def apply_pow_to_dims(dim_col: Sequence[int | float], pwr: float) -> List[int]:
    """*apply_pow_to_dims()* Apply power operation to dimensional column.

    Args:
        dim_col (Sequence[int | float]): Dimensional column.
        pwr (float): Power to apply.

    Returns:
        List[int]: Dimensional column with power applied.

    Example:
        >>> apply_pow_to_dims([1, 2, -1], -1)
        [-1, -2, 1]  # (L^1 * M^2 * T^-1)^-1 = L^-1 * M^-2 * T^1
    """
    return [int(exp * pwr) for exp in dim_col]


def validate_coeff_expr(expr: str,
                        avail_coeffs: Dict[str, Any],
                        pattern: str = PI_COEF_RE) -> None:
    """*validate_coeff_expr()* Validate coefficient expression.

    Checks that:
        1. Expression contains valid coefficient references
        2. All referenced coefficients exist

    Args:
        expr (str): Expression to validate.
        avail_coeffs (Dict[str, Any]): Available coefficients.
        pattern (str): Regex pattern to match coefficients.

    Raises:
        ValueError: If expression is invalid or references missing coefficients.
    """
    # Extract coefficient symbols
    coef_symbols = extract_coeff_syms(expr, pattern)

    if not coef_symbols:
        raise ValueError(
            f"Expression '{expr}' does not contain any valid "
            f"coefficient references (format: \\Pi_{{n}})."
        )

    # Check all coefficients exist
    for sym in coef_symbols:
        if sym not in avail_coeffs:
            raise ValueError(f"Referenced coefficient {sym} does not exist.")


def parse_dim_expr(expr: str,
                   coeffs: Dict[str, Any],
                   dim_col_fn: Callable[[Any], List[float | int]],
                   coef_patt: str = PI_COEF_RE,
                   pow_patt: str = PI_POW_RE,
                   ops_patt: str = BASIC_OPS_RE) -> Tuple[List[int], Optional[float]]:
    """*parse_dim_expr()* Parse and evaluate dimensional expression.

    Main parsing function that processes a dimensional expression containing
    Pi coefficients, operations, and numeric constants.

    Args:
        expr (str): Mathematical expression to parse.
        coeffs (Dict[str, Any]): Available coefficients dictionary.
        dim_col_fn (Callable[[Any], List[float]]): Function to extract dimension column from coefficient. Signature: dim_col_fn(coef) -> List[float]
        coef_patt (str): Regex pattern to match coefficients.
        pow_patt (str): Regex pattern to match power expressions.
        ops_patt (str): Regex pattern to match operations.

    Returns:
        Tuple[List[int], Optional[float]]:
            - Resulting dimensional column
            - Numeric constant multiplier (None if no constant or constant = 1.0)

    Raises:
        ValueError: If expression parsing fails.

    Example:
        >>> def get_dims(c): return c._dim_col
        >>> result_dims, constant = parse_dim_expr(
        ...     "2 * \\Pi_{0}**(-1) * \\Pi_{1}",
        ...     coeffs,
        ...     get_dims
        ... )
    """
    # Validate expression first
    validate_coeff_expr(expr, coeffs, coef_patt)

    # Extract powered coefficients and replace with placeholders
    expr, pow_transforms = extract_powered_coeffs(expr, pow_patt)

    # Parse expression into tokens
    parts = re.split(ops_patt, expr)

    # Initialize state
    _cur_op = "*"
    _res_dim_col: List[int] = []
    numeric_constant: Optional[float] = None
    is_first = True

    for part in parts:
        part = part.strip()

        # Skip empty parts
        if not part:
            continue

        # Check if it's an operator
        if part in ("*", "/", "^", "+", "-"):
            _cur_op = part
            continue

        # Check if it's a placeholder for powered coefficient
        if part.startswith("__PHLR_"):
            coef_sym, power = pow_transforms[part]
            coef = coeffs[coef_sym]

            # Get and apply power to dimensions
            dim_col = dim_col_fn(coef)
            _pow_dims = apply_pow_to_dims(dim_col, power)

            if is_first:
                _res_dim_col = _pow_dims
                is_first = False
            else:
                _res_dim_col = compute_dims_opn(_res_dim_col,
                                                _pow_dims,
                                                _cur_op)

        # Check if it's a numeric constant
        elif re.match(r"^[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?$", part):
            constant = float(part)

            # Initialize with zeros if first element
            if is_first and not _res_dim_col:
                # Get dimension count from first coefficient
                coef_symbols = extract_coeff_syms(expr, coef_patt)
                if coef_symbols:
                    ref_coef = coeffs[coef_symbols[0]]
                    _res_dim_col = [0] * len(dim_col_fn(ref_coef))
                    is_first = False

            # Track numeric constants
            if _cur_op == "*":
                numeric_constant = constant if numeric_constant is None else numeric_constant * constant
            elif _cur_op == "/":
                numeric_constant = 1.0 / constant if numeric_constant is None else numeric_constant / constant

        # Check if it's a regular coefficient
        elif re.match(coef_patt, part):
            coef = coeffs[part]
            _dim_col = dim_col_fn(coef)

            if is_first:
                _res_dim_col = [int(x) for x in _dim_col]
                is_first = False
            else:
                _res_dim_col = compute_dims_opn(_res_dim_col,
                                                _dim_col,
                                                _cur_op)

    # Validate result
    if not _res_dim_col:
        _msg = f"Failed to derive dimensions from expression '{expr}'. "
        _msg += "Expression must contain valid coefficient references."
        raise ValueError(_msg)
    # Return result
    return _res_dim_col, numeric_constant


def format_numeric_constant(constant: Optional[float]) -> str:
    """*format_numeric_constant()* Format numeric constant for display.

    Args:
        constant (Optional[float]): Numeric constant to format.

    Returns:
        str: Formatted constant string, empty if None or 1.0.

    Example:
        >>> format_numeric_constant(2.0)
        '2'
        >>> format_numeric_constant(0.5)
        '0.5'
        >>> format_numeric_constant(None)
        ''
    """
    if constant is None or constant == 1.0:
        return ""

    if constant == int(constant):
        return str(int(constant))

    return str(constant)



# -*- coding: utf-8 -*-
"""
Module config.py
===========================================

Configuration module for basic *PyDASA* configuration parameters related to Dimensional Analysis (DA) of complex phenomena.

This module provides type-safe configuration through Enums and frozen dataclasses, replacing the previous mutable dictionary-based approach with immutable, type-checked alternatives.

NOTE: in the future the enum should be configurated via external files (e.g., JSON, YAML) to allow user customization.

Key Features:
    - Type-safe Enum definitions for frameworks, categories, and modes
    - Immutable configuration via frozen dataclass with singleton pattern
    - Backward compatibility with legacy dict-based access

    Supported Frameworks:
        - PHYSICAL: Traditional physical dimensional framework.
        - COMPUTATION: Computer science dimensional framework.
        - SOFTWARE: Software architecture dimensional framework.
        - CUSTOM: User-defined dimensional framework.

    Supported Variable Categories:
        - IN: Input variables influencing the system.
        - OUT: Output variables representing analysis results.
        - CTRL: Control variables constraining the system.

    Supported Analysis Modes:
        - SYM: Analysis for symbolic processable Parameters (e.g., 'x + y').
        - NUM: Analysis for numeric Variables (e.g., 1.0, 2.5).

*IMPORTANT* Based on:
    # H.Gorter, *Dimensionalanalyse: Eine Theoririe der physikalischen Dimensionen mit Anwendungen*
"""
# native python modules
from enum import Enum
# dataclass imports
from dataclasses import dataclass, field
# data type imports
from typing import ClassVar

# custom modules
from pydasa.core.io import Path, load     # , save
from pydasa.core.constants import DFLT_CFG_FOLDER, DFLT_CFG_FILE

# checking custom modules
assert load
assert DFLT_CFG_FOLDER
assert DFLT_CFG_FILE

# =============================================================================
# *PyDASA* Enum Definitions
# =============================================================================


class Frameworks(str, Enum):
    """**Frameworks** Enum for Fundamental Dimensional Units (FDUs) frameworks in *PyDASA*.

    Purpose:
        - Defines the dimensional frameworks supported in *PyDASA*.

    Args:
        str (class): Python native str class.
        Enum (class): Python native Enum class.

    Returns:
        Frameworks: Enum member representing the FDU framework.
    """

    PHYSICAL = "PHYSICAL"
    COMPUTATION = "COMPUTATION"
    SOFTWARE = "SOFTWARE"
    CUSTOM = "CUSTOM"

    @property
    def description(self) -> str:
        """*description* Get human-readable description of the framework.

        Returns:
            str: Human-readable framework's description.
        """

        descriptions = {
            Frameworks.PHYSICAL: "Traditional physical dimensional framework (e.g., Length, Mass, Time).",
            Frameworks.COMPUTATION: "Computer science dimensional framework (e.g., Time, Space, Complexity).",
            Frameworks.SOFTWARE: "Software architecture dimensional framework (e.g., Time, Data, Connectivity).",
            Frameworks.CUSTOM: "User-defined dimensional framework for specific use cases.",
        }
        return descriptions[self]


class VarCardinality(str, Enum):
    """**VarCardinality** Enum for Variable cardinality in *PyDASA*.

    Purpose:
        - Defines the variable categories supported in *PyDASA*.
        - Used to classify variables in the dimensional matrix.

    Args:
        str (class): Python native str class.
        Enum (class): Python native Enum class.

    Returns:
        VarCardinality: Enum member representing the variable cardinality.
    """
    IN = "IN"
    OUT = "OUT"
    CTRL = "CTRL"

    @property
    def description(self) -> str:
        """*description* Get human-readable description of the variable cardinality.

        Returns:
            str: Human-readable variable cardinality description.
        """
        descriptions = {
            VarCardinality.IN: "Variables that influence the system (e.g., known inputs).",
            VarCardinality.OUT: "Variable that represent the results of the analysis.",
            VarCardinality.CTRL: "Variables used to control or constrain the system (e.g., constants).",
        }
        return descriptions[self]


class CoefCardinality(str, Enum):
    """**CoefCardinality** Enum for Dimensionless Coefficient/Numbers/Groups (DC/DN/DG) cardinality in *PyDASA*.

    Purpose:
        - Defines the categories of dimensionless coefficients supported in *PyDASA*.
        - Used to classify dimensionless coefficients in formulas and equations.
        - Helps in organizing and managing dimensionless coefficients in the analysis.

    Args:
        str (class): Python native str class.
        Enum (class): Python native Enum class.

    Returns:
        CoefCardinality: Enum member representing the coefficient cardinality.
    """
    COMPUTED = "COMPUTED"
    DERIVED = "DERIVED"

    @property
    def description(self) -> str:
        """*description* Get human-readable description of the coefficient cardinality.

        Returns:
            str: Human-readable coefficient cardinality description.
        """
        descriptions = {
            CoefCardinality.COMPUTED: "Coefficients directly calculated using the Dimensional Matrix.",
            CoefCardinality.DERIVED: "Coefficients obtained by combining or manipulating Computed Coefficients.",
        }
        return descriptions[self]


class AnaliticMode(str, Enum):
    """**AnaliticMode** Enum for analysis modes (e.g. sensitivity analysis, Monte Carlo simulation) in *PyDASA*.

    Purpose:
        - Defines the analysis modes supported in *PyDASA*.
        - Used to specify the type of analysis to be performed on variables, coefficients, or functions.

    Args:
        str (class): Python native str class.
        Enum (class): python native Enum class.

    Returns:
        AnaliticMode: Enum member representing the analysis mode.
    """
    SYM = "SYM"
    NUM = "NUM"

    @property
    def description(self) -> str:
        descriptions = {
            AnaliticMode.SYM: "Analysis for symbolic processable parameters (e.g., 'z = x + y').",
            AnaliticMode.NUM: "analysis for numeric variable ranges (e.g., 1.0, 2.5).",
        }
        return descriptions[self]


class SimulationMode(str, Enum):
    """**SimulationMode** Enum for simulation modes in *PyDASA*.

    Attributes:
        DIST: Generate samples using distribution functions (Monte Carlo).
        DATA: Use pre-existing data from Variable._data attributes.
    """
    DIST = "DIST"
    DATA = "DATA"

    @property
    def description(self) -> str:
        descriptions = {
            SimulationMode.DIST: "Generate samples using distribution functions (Monte Carlo).",
            SimulationMode.DATA: "Use pre-existing data from Variable._data attributes.",
        }
        return descriptions[self]

# =============================================================================
# Immutable Configuration Singleton
# =============================================================================


@dataclass(frozen=True)
class PyDASAConfig:
    """ **PyDASAConfig** Singleton class for PyDASA configuration. It uses dataclass decorator to freeze the data.

    Returns:
        PyDASAConfig: frozen singleton configuration instance.
    """

    # :attr: _instance
    _instance: ClassVar["PyDASAConfig | None"] = None
    """Singleton instance of PyDASAConfig."""

    # :attr: SPT_FDU_FWKS
    SPT_FDU_FWKS: dict = field(default_factory=dict)
    """Supported Fundamental Dimensional Units (FDUs) frameworks and their configurations."""

    def __post_init__(self):
        """*__post_init__()* Post-initialization to load configuration from file."""
        # Load configuration from default file (relative to this module's directory)
        module_dir = Path(__file__).parent
        fp = module_dir / DFLT_CFG_FOLDER / DFLT_CFG_FILE
        cfg_data = load(fp)

        # Since the dataclass is frozen, use object.__setattr__ to set attributes
        object.__setattr__(self,
                           "SPT_FDU_FWKS",
                           cfg_data.get("frameworks", {}))

    @classmethod
    def get_instance(cls) -> "PyDASAConfig":
        """*get_instance()* Get the singleton instance of PyDASAConfig.

        Returns:
            PyDASAConfig: Singleton instance of PyDASAConfig.
        """
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @property
    def frameworks(self) -> tuple[Frameworks, ...]:
        """*frameworks* Get supported frameworks.

        Returns:
            tuple[Frameworks, ...]: Tuple of supported Frameworks.
        """
        return tuple(Frameworks)

    @property
    def parameter_cardinality(self) -> tuple[VarCardinality, ...]:
        """*parameter_cardinality* Get supported variable cardinalities.

        Returns:
            tuple[VarCardinality, ...]: Tuple of supported VarCardinality.
        """
        return tuple(VarCardinality)

    @property
    def coefficient_cardinality(self) -> tuple[CoefCardinality, ...]:
        """*coefficient_cardinality* Get supported coefficient cardinalities.

        Returns:
            tuple[CoefCardinality, ...]: Tuple of supported CoefCardinality.
        """
        return tuple(CoefCardinality)

    @property
    def analitic_modes(self) -> tuple[AnaliticMode, ...]:
        """*analitic_modes* Get supported analysis modes.

        Returns:
            tuple[AnaliticMode, ...]: Tuple of supported AnaliticMode.
        """
        return tuple(AnaliticMode)

    @property
    def simulation_modes(self) -> tuple[SimulationMode, ...]:
        """*simulation_modes* Get supported simulation modes.

        Returns:
            tuple[SimulationMode, ...]: Tuple of supported SimulationMode.
        """
        return tuple(SimulationMode)


# Get singleton instance for configuration
# :attr: PYDASA_CFG
PYDASA_CFG: PyDASAConfig = PyDASAConfig()
"""
Singleton instance of PyDASAConfig for accessing global configuration.
"""



# -*- coding: utf-8 -*-
"""
Module serialization
===========================================

Serialization and parsing utilities for PyDASA.

This module provides utilities for:
- LaTeX expression parsing and conversion
- Symbol mapping and transformation
- Future: JSON, YAML serialization support

Modules:
    **parser**: LaTeX parsing and Python conversion utilities
"""

from pydasa.serialization.parser import (
    latex_to_python,
    parse_latex,
    create_latex_mapping,
)

__all__ = [
    "latex_to_python",
    "parse_latex", 
    "create_latex_mapping",
]



# -*- coding: utf-8 -*-
"""
PyDASA Package
===========================================
Main package initializer for PyDASA library, exposing core functionalities, data structures, dimensional analysis tools, unit conversion utilities, and workflow management components.
"""
# expose version
from ._version import __version__

# expose imports
# exposing analytics modules
from .analysis.scenario import Sensitivity
from .analysis.simulation import MonteCarlo

# TODO conversion still in development
# exposing unit conversion modules
# from .context.conversion import UnitStandarizer
# from .context.system import MeasureSystem
# from .context.measurements import Unit

# exposing pi-theorem/dimensional analysis modules
from .dimensional.buckingham import Coefficient
from .dimensional.fundamental import Dimension
from .dimensional.vaschy import Schema
from .dimensional.model import Matrix

# exposing core modules
# exposing basic elements/variables modules
from .elements.parameter import Variable
# exposing parser/io modules
from .core.io import load, save

# exposing custom data structure modules
# TODO measurement still in development
# lists
from .structs.lists.arlt import ArrayList
from .structs.lists.sllt import SingleLinkedList
from .structs.lists.ndlt import Node, SLNode, DLNode
# from .structs.lists.dllt import DoubleLinkedList
# tables
from .structs.tables.htme import MapEntry
from .structs.tables.scht import Bucket, SCHashTable

# exposing validation, error and decorator modules
# exposing workflow modules
from .workflows.basic import WorkflowBase
from .workflows.influence import SensitivityAnalysis
from .workflows.practical import MonteCarloSimulation
from .workflows.phenomena import AnalysisEngine

# asserting all imports
# asserting analytics modules
assert Sensitivity
assert MonteCarlo

# asserting unit conversion modules
# assert UnitStandarizer
# assert MeasureSystem
# assert Unit
# asserting pi-theorem/dimensional analysis modules
assert Coefficient
assert Dimension
assert Schema
assert Matrix
# asserting elements/variables modules
assert Variable
# asserting parser/io modules
assert load
assert save
# asserting custom data structure modules
# lists
assert ArrayList
assert SingleLinkedList
assert Node
assert SLNode
assert DLNode
# tables
assert MapEntry
assert Bucket
assert SCHashTable
# asserting workflow modules
assert WorkflowBase
assert SensitivityAnalysis
assert MonteCarloSimulation
assert AnalysisEngine

# Define __all__ for wildcard imports
__all__ = [
    "__version__",
    "Sensitivity",
    "MonteCarlo",
    # "UnitStandarizer",
    # "MeasureSystem",
    # "Unit",
    "Coefficient",
    "Dimension",
    "Schema",
    "Matrix",
    "Variable",
    "load",
    "save",
    "WorkflowBase",
    "SensitivityAnalysis",
    "MonteCarloSimulation",
    "AnalysisEngine",
]



# -*- coding: utf-8 -*-
"""
Module numerical.py
===========================================

Numerical perspective for variable representation.

This module defines the NumericalSpecs class representing the computational
value ranges and discretization properties of a variable.

Classes:
    **BoundsSpecs**: Value bounds in original units
    **StandardizedSpecs**: Standardized value specifications with discretization
    **NumericalSpecs**: Numerical variable specifications (combines BoundsSpecs and StandardizedSpecs)

*IMPORTANT:* Based on the theory from:

    # H.Gorter, *Dimensionalanalyse: Eine Theoririe der physikalischen Dimensionen mit Anwendungen*
"""
# native python modules
# dataclass imports
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional

# third-party numerical imports
import numpy as np
from numpy.typing import NDArray

# custom modules
from pydasa.validations.decorators import validate_type
from pydasa.validations.decorators import validate_range


@dataclass
class BoundsSpecs:
    """Value bounds in original units: min, max, mean, deviation, setpoint.

    This class represents the value ranges and statistical measures
    in the original (non-standardized) units of a variable.

    Attributes:
        _setpoint (Optional[float]): Specific value/point in original units.
        _min (Optional[float]): Minimum value in original units.
        _max (Optional[float]): Maximum value in original units.
        _mean (Optional[float]): Mean value in original units.
        _median (Optional[float]): Median value in original units.
        _dev (Optional[float]): Standard deviation in original units.
    """

    # Value ranges (original units)
    # :attr: _setpoint
    _setpoint: Optional[float] = None
    """Specific value in original units."""

    # :attr: _min
    _min: Optional[float] = None
    """Minimum value in original units."""

    # :attr: _max
    _max: Optional[float] = None
    """Maximum value in original units."""

    # :attr: _mean
    _mean: Optional[float] = None
    """Mean value in original units."""

    # :attr: _median
    _median: Optional[float] = None
    """Median value in original units."""

    # :attr: _dev
    _dev: Optional[float] = None
    """Standard deviation in original units."""

    @property
    def setpoint(self) -> Optional[float]:
        """*setpoint* Get specific value/point system for original units.

        Returns:
            Optional[str]: Specific value/point system for original units.
        """
        return self._setpoint

    @setpoint.setter
    @validate_type(int, float, allow_none=False, allow_nan=True)
    def setpoint(self, val: Optional[float]) -> None:
        """*setpoint* Sets specific value/point system for original units.

        Args:
            val (Optional[float]): Specific value/point system for original units.

        Raises:
            ValueError: If value not a valid number.
        """
        self._setpoint = val

    @property
    def min(self) -> Optional[float]:
        """*min* Get minimum range value.

        Returns:
            Optional[float]: Minimum range value.
        """
        return self._min

    @min.setter
    @validate_type(int, float, allow_none=False, allow_nan=True)
    @validate_range(max_attr="_max")
    def min(self, val: Optional[float]) -> None:
        """*min* Sets minimum range value.

        Args:
            val (Optional[float]): Minimum range value.

        Raises:
            ValueError: If value not a valid number.
            ValueError: If value is greater than max.
        """
        self._min = val

    @property
    def max(self) -> Optional[float]:
        """*max* Get the maximum range value.

        Returns:
            Optional[float]: Maximum range value.
        """
        return self._max

    @max.setter
    @validate_type(int, float, allow_none=False, allow_nan=True)
    @validate_range(min_attr="_min")
    def max(self, val: Optional[float]) -> None:
        """*max* Sets the maximum range value.

        Args:
            val (Optional[float]): Maximum range value.

        Raises:
            ValueError: If value is not a valid number.
            ValueError: If value is less than min.
        """
        self._max = val

    @property
    def mean(self) -> Optional[float]:
        """*mean* Get the Variable average value.

        Returns:
            Optional[float]: Variable average value.
        """
        return self._mean

    @mean.setter
    @validate_type(int, float, allow_none=False, allow_nan=True)
    @validate_range(min_attr="_min", max_attr="_max")
    def mean(self, val: Optional[float]) -> None:
        """*mean* Sets the Variable mean value.

        Args:
            val (Optional[float]): Variable mean value.

        Raises:
            ValueError: If value not a valid number.
            ValueError: If value is outside min-max range.
        """
        self._mean = val

    @property
    def median(self) -> Optional[float]:
        """*median* Get the median value.

        Returns:
            Optional[float]: Median value.
        """
        return self._median

    @median.setter
    @validate_type(int, float, allow_none=False, allow_nan=True)
    @validate_range(min_attr="_min", max_attr="_max")
    def median(self, val: Optional[float]) -> None:
        """*median* Sets the median value.

        Args:
            val (Optional[float]): Median value.

        Raises:
            ValueError: If value not a valid number.
            ValueError: If value is outside min-max range.
        """
        self._median = val

    @property
    def dev(self) -> Optional[float]:
        """*dev* Get the Variable standardized deviation.

        Returns:
            Optional[float]: Variable standardized deviation.
        """
        return self._dev

    @dev.setter
    @validate_type(int, float, allow_none=False, allow_nan=True)
    @validate_range(min_value=0, min_inclusive=True)
    def dev(self, val: Optional[float]) -> None:
        """*dev* Sets the Variable standardized deviation.

        Args:
            val (Optional[float]): Variable standardized deviation.
        Raises:
            ValueError: If value not a valid number.
            ValueError: If value is negative.
        """
        # if val is not None and not isinstance(val, (int, float)):
        #     raise ValueError("Standard deviation must be a number.")

        self._dev = val

    def clear(self) -> None:
        """*clear()* Reset bounds attributes to default values.

        Resets all value ranges in original units.
        """
        self._setpoint = None
        self._min = None
        self._max = None
        self._mean = None
        self._median = None
        self._dev = None


@dataclass
class StandardizedSpecs:
    """Standardized value specifications: ranges, statistics, and discretization.

    This class represents value ranges and statistical measures in standardized units,
    plus discretization properties for numerical analysis (step size, range array).

    Attributes:
        _std_setpoint (Optional[float]): Specific value/point in standardized units.
        _std_min (Optional[float]): Minimum value in standardized units.
        _std_max (Optional[float]): Maximum value in standardized units.
        _std_mean (Optional[float]): Mean value in standardized units.
        _std_median (Optional[float]): Median value in standardized units.
        _std_dev (Optional[float]): Standard deviation in standardized units.
    """

    # Value ranges (standardized units)
    # :attr: _std_setpoint
    _std_setpoint: Optional[float] = None
    """Specific value/point system for standardized units."""

    # :attr: _std_min
    _std_min: Optional[float] = None
    """Minimum value in standardized units."""

    # :attr: _std_max
    _std_max: Optional[float] = None
    """Maximum value in standardized units."""

    # :attr: _std_mean
    _std_mean: Optional[float] = None
    """Mean value in standardized units."""

    # :attr: _std_median
    _std_median: Optional[float] = None
    """Median value in standardized units."""

    # :attr: _std_dev
    _std_dev: Optional[float] = None
    """Standard deviation in standardized units."""

    @property
    def std_setpoint(self) -> Optional[float]:
        """*std_setpoint* Get specific value/point system for standardized units.

        Returns:
            Optional[str]: Specific value/point system for standardized units.
        """
        return self._std_setpoint

    @std_setpoint.setter
    @validate_type(int, float, allow_none=False)
    def std_setpoint(self, val: Optional[float]) -> None:
        """*std_setpoint* Sets specific value/point system for standardized units.

        Args:
            val (Optional[float]): Specific value/point system for standardized units.
        Raises:
            ValueError: If value not a valid number.
        """
        self._std_setpoint = val

    @property
    def std_min(self) -> Optional[float]:
        """*std_min* Get the standardized minimum range value.

        Returns:
            Optional[float]: standardized minimum range value.
        """
        return self._std_min

    @std_min.setter
    @validate_type(int, float, allow_none=False)
    @validate_range(max_attr="_std_max")
    def std_min(self, val: Optional[float]) -> None:
        """*std_min* Sets the standardized minimum range value.

        Args:
            val (Optional[float]): standardized minimum range value.

        Raises:
            ValueError: If value not a valid number.
            ValueError: If value is greater than std_max.
        """
        self._std_min = val

    @property
    def std_max(self) -> Optional[float]:
        """*std_max* Get the standardized maximum range value.

        Returns:
            Optional[float]: standardized maximum range value.
        """
        return self._std_max

    @std_max.setter
    @validate_type(int, float, allow_none=False)
    @validate_range(min_attr="_std_min")
    def std_max(self, val: Optional[float]) -> None:
        """*std_max* Sets the standardized maximum range value.

        Raises:
            ValueError: If value is not a valid number.
            ValueError: If value is less than std_min.
        """
        self._std_max = val

    @property
    def std_mean(self) -> Optional[float]:
        """*std_mean* Get standardized mean value.

        Returns:
            Optional[float]: standardized mean.
        """
        return self._std_mean

    @std_mean.setter
    @validate_type(int, float, allow_none=False)
    @validate_range(min_attr="_std_min", max_attr="_std_max")
    def std_mean(self, val: Optional[float]) -> None:
        """*std_mean* Sets the standardized mean value.

        Args:
            val (Optional[float]): standardized mean value.

        Raises:
            ValueError: If value is not a valid number.
            ValueError: If value is outside std_min-std_max range.
        """
        self._std_mean = val

    @property
    def std_median(self) -> Optional[float]:
        """*std_median* Get standardized median value.

        Returns:
            Optional[float]: Standardized median.
        """
        return self._std_median

    @std_median.setter
    @validate_type(int, float, allow_none=False)
    @validate_range(min_attr="_std_min", max_attr="_std_max")
    def std_median(self, val: Optional[float]) -> None:
        """*std_median* Sets the standardized median value.

        Args:
            val (Optional[float]): Standardized median value.

        Raises:
            ValueError: If value is not a valid number.
            ValueError: If value is outside std_min-std_max range.
        """
        self._std_median = val

    @property
    def std_dev(self) -> Optional[float]:
        """*std_dev* Get standardized standardized deviation.

        Returns:
            Optional[float]: Standardized standardized deviation.
        """
        return self._std_dev

    @std_dev.setter
    @validate_type(int, float, allow_none=False)
    @validate_range(min_value=0, min_inclusive=True)
    def std_dev(self, val: Optional[float]) -> None:
        """*std_dev* Sets the standardized standardized deviation.

        Args:
            val (Optional[float]): Standardized standardized deviation.

        Raises:
            ValueError: If value is not a valid number.
            ValueError: If value is negative.
        """
        self._std_dev = val

    def clear(self) -> None:
        """*clear()* Reset standardized attributes to default values.

        Resets all value ranges in standardized units.
        """
        self._std_setpoint = None
        self._std_min = None
        self._std_max = None
        self._std_mean = None
        self._std_median = None
        self._std_dev = None


@dataclass
class NumericalSpecs(BoundsSpecs, StandardizedSpecs):
    """Numerical perspective: computational value ranges. Answers the question: "What VALUES can this variable take?"

    This perspective combines:
        - BoundsSpecs: Value ranges in original units (min, max, mean, deviation, setpoint)
        - StandardizedSpecs: Value ranges in standardized units
        - Discretization properties (step, range array)

    This perspective focuses on:
        - Concrete bounds (minimum, maximum)
        - Central tendency (mean value)
        - Variation (standard deviation)
        - Discretization for simulations (step size, range arrays)
        - Unit conversions (original <-> standardized)
        - Variable dependencies (calculated variables)

    Attributes:
        # From BoundsSpecs (original units):
            _setpoint (Optional[float]): Specific value/point in original units.
            _min (Optional[float]): Minimum value in original units.
            _max (Optional[float]): Maximum value in original units.
            _mean (Optional[float]): Mean value in original units.
            _median (Optional[float]): Median value in original units.
            _dev (Optional[float]): Standard deviation in original units.
        # From StandardizedSpecs (standardized units):
            _std_setpoint (Optional[float]): Specific value/point in standardized units.
            _std_min (Optional[float]): Minimum value in standardized units.
            _std_max (Optional[float]): Maximum value in standardized units.
            _std_mean (Optional[float]): Mean value in standardized units.
            _std_median (Optional[float]): Median value in standardized units.
            _std_dev (Optional[float]): Standard deviation in standardized units.
        # Discretization properties:
            _step (Optional[float]): Step size for simulations.
            _data (NDArray[np.float64]): Range for numerical analysis in standardized units.
    """

    # :attr: _step
    _step: Optional[float] = 1e-3
    """Step size for simulations."""

    # :attr: _data
    _data: NDArray[np.float64] = field(
        default_factory=lambda: np.array([], dtype=np.float64)
    )
    """Data range array for analysis."""

    @property
    def step(self) -> Optional[float]:
        """*step* Get standardized step size.

        Returns:
            Optional[float]: Step size (always standardized).
        """
        return self._step

    @step.setter
    @validate_type(int, float, allow_none=False)
    @validate_range(min_value=0,
                    min_inclusive=False,
                    min_attr="_std_min",
                    max_attr="_std_max")
    def step(self, val: Optional[float]) -> None:
        """*step* Set standardized step size.

        Args:
            val (Optional[float]): Step size (always standardized).
        """
        self._step = val

    @property
    def data(self) -> NDArray[np.float64]:
        """*data* Get standardized data array.

        Returns:
            NDArray[np.float64]: Data array for range (always standardized).
        """
        return self._data

    @data.setter
    @validate_type(list, np.ndarray, allow_none=False)
    def data(self, val: NDArray[np.float64]) -> None:
        """*data* Set standardized data array.

        Args:
            val (NDArray[np.float64]): Data array for range (always standardized).

        Raises:
            ValueError: If value is not a numpy array or list.
        """
        # always store as numpy array
        if isinstance(val, list):
            val = np.array(val, dtype=np.float64)
        self._data = val

    def generate_data(self) -> None:
        """*generate_data()* Generate standardized data array from min, max, using step value.

        Raises:
            ValueError: If needed values are missing.
        """
        # Type narrow using local variables
        std_min = self._std_min
        std_max = self._std_max
        step = self._step

        if std_min is not None and std_max is not None and step is not None:
            self._data = np.arange(std_min, std_max, step)
        else:
            _msg = "Cannot generate data array. Needed values are missing: "
            _msg += f"std_min={std_min}, std_max={std_max}, step={step}."
            raise ValueError(_msg)

    def clear(self) -> None:
        """*clear()* Reset numerical attributes to default values.

        Resets all value ranges, discretization, and step size by calling parent clear() methods.
        """
        BoundsSpecs.clear(self)
        StandardizedSpecs.clear(self)
        self._step = None
        self._data = np.array([])



# -*- coding: utf-8 -*-
"""
Variable specifications module.

Four compositional perspectives for variable representation:
- ConceptualSpecs: Identity and classification
- SymbolicSpecs: Mathematical notation and dimensions
- NumericalSpecs: Computational values and ranges
- StatisticalSpecs: Probabilistic distributions
"""

from .conceptual import ConceptualSpecs
from .symbolic import SymbolicSpecs
from .numerical import NumericalSpecs
from .statistical import StatisticalSpecs

__all__ = [
    'ConceptualSpecs',
    'SymbolicSpecs',
    'NumericalSpecs',
    'StatisticalSpecs',
]



# -*- coding: utf-8 -*-
"""
Module basic.py
===========================================

This module provides base classes with common validation logic used across FDU, Parameter, and Variable classes to eliminate code duplication.

Classes:
    **Foundation**: enforces common validation logic.
    **IdxBasis**: enforces index/precedence validation logic.
    **SymBasis**: enforces symbol and framework validation logic.

*IMPORTANT:* Based on the theory from:

    # H.Gorter, *Dimensionalanalyse: Eine Theoririe der physikalischen Dimensionen mit Anwendungen*
"""
# native python modules
# forward references + postpone eval type hints
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
import inspect

# indicate it is an abstract base class
from abc import ABC

# import global variables
from pydasa.core.setup import Frameworks
from pydasa.core.setup import PYDASA_CFG
from pydasa.validations.patterns import LATEX_RE

# import validation decorators
from pydasa.validations.decorators import validate_type
from pydasa.validations.decorators import validate_emptiness
from pydasa.validations.decorators import validate_choices
from pydasa.validations.decorators import validate_pattern
from pydasa.validations.decorators import validate_index


@dataclass
class SymBasis(ABC):
    """**SymBasis** Foundation class for entities with symbols and framework functionalities.

    Args:
        ABC (ABC): Abstract base class.

    Attributes:
        _sym (str): Symbol representation.
        _alias (str): Python-compatible alias for use in code.
        _fwk (str): Frameworks context.
    """

    # :attr: _sym
    _sym: str = ""
    """Symbol representation (LaTeX or alphanumeric)."""

    # :attr: _fwk
    _fwk: str = Frameworks.PHYSICAL.value
    """Frameworks context (PHYSICAL, COMPUTATION, SOFTWARE, CUSTOM)."""

    # :attr: _alias
    _alias: str = ""
    """Python-compatible alias for symbol, used in executable code. e.g.: `\\rho_{1}` -> `rho_1`."""

    def __post_init__(self) -> None:
        """*__post_init__()* Post-initialization processing with symbol and framework validation.
        """
        # super().__post_init__()
        # Validate the symbol and framework
        if not self._sym:
            self._sym = self._sym.strip()
        if not self._fwk:
            self._fwk = self._fwk.strip()
        if not self._alias:
            self._alias = self._alias.strip()

    @property
    def sym(self) -> str:
        """*sym* Get the symbol.

        Returns:
            str: Symbol value.
        """
        return self._sym

    @sym.setter
    @validate_type(str)
    @validate_emptiness()
    @validate_pattern(LATEX_RE, allow_alnum=True)
    def sym(self, val: str) -> None:
        """*sym* Set the symbol with validation.

        Args:
            val (str): Symbol value.

        Raises:
            ValueError: If symbol format is invalid.
        """
        self._sym = val

    @property
    def fwk(self) -> str:
        """*fwk* Get the framework.

        Returns:
            str: Frameworks value.
        """
        return self._fwk

    @fwk.setter
    @validate_type(str)
    @validate_choices(PYDASA_CFG.frameworks)
    def fwk(self, val: str) -> None:
        """*fwk* Set the framework with validation.

        Args:
            val (str): Frameworks value.

        Raises:
            ValueError: If framework is not supported.
        """
        self._fwk = val

    @property
    def alias(self) -> Optional[str]:
        """*alias* Get the Python variable synonym.

        Returns:
            Optional[str]: Python variable name. e.g.: `\\rho_{1}` -> `rho_1`.
        """
        return self._alias

    @alias.setter
    @validate_type(str)
    @validate_emptiness()
    def alias(self, val: str) -> None:
        """*alias* Set the Python variable synonym.

        Args:
            val (str): Python variable name. e.g.: `\\rho_{1}` -> `rho_1`.

        Raises:
            ValueError: If variable name is empty.
        """
        self._alias = val

    def clear(self) -> None:
        """*clear()* Reset symbol and framework attributes to default values.

        Resets the entity's symbol-related properties to their initial state.
        """
        # Reset symbol attributes
        self._sym = ""
        self._alias = ""
        self._fwk = Frameworks.PHYSICAL.value


@dataclass
class IdxBasis(SymBasis):
    """**IdxBasis** Foundation class for entities with index/precedence functionality.

    Args:
        SymBasis (SymBasis): Inherits symbol and framework validation.

    Attributes:
        _idx (int): Index/precedence value
    """

    # :attr: _idx
    _idx: int = -1
    """Unique identifier/index for ordering in dimensional matrix."""

    def __post_init__(self) -> None:
        """*__post_init__()* Post-initialization processing with symbol and framework validation.
        """
        super().__post_init__()
        if self._idx != -1:
            self.idx = self._idx

    @property
    def idx(self) -> int:
        """*idx* Get the index/precedence value.

        Returns:
            int: Index value.
        """
        return self._idx

    @idx.setter
    @validate_type(int, allow_none=False)
    @validate_index()
    def idx(self, val: int) -> None:
        """*idx* Set the index/precedence value.

        Args:
            val (int): Index value (must be non-negative).

        Raises:
            ValueError: If index is not a non-negative integer.
        """
        self._idx = val

    def clear(self) -> None:
        """*clear()* Reset index and inherited attributes to default values.

        Resets the entity's index and symbol-related properties to their initial state.
        """
        # Reset parent class attributes
        super().clear()

        # Reset index attribute
        self._idx = -1


@dataclass
class Foundation(IdxBasis):
    """**Foundation** Foundation class for all dimensional analysis entities.

    Provides common validation logic and attributes shared by FDU, Variable, and Coeffcient classes.

    Args:
        IdxBasis (IdxBasis): Inherits index, symbol, and framework validation.

    Attributes:
        name (str): User-friendly name
        description (str): Brief summary or description
    """

    # :attr: _name
    _name: str = ""
    """User-friendly name of the entity."""

    # :attr: description
    description: str = ""
    """Brief summary or description of the entity."""

    def __post_init__(self) -> None:
        """*__post_init__()* Post-initialization processing with description capitalization.
        """
        if self.description:
            self.description = self.description.strip()

    @property
    def name(self) -> str:
        """*name* Get the name.

        Returns:
            str: Name value.
        """
        return self._name

    @name.setter
    @validate_type(str, allow_none=False)
    def name(self, val: str) -> None:
        """*name* Set the name with validation.

        Args:
            val (str): Name value.

        Raises:
            ValueError: If name is not a non-empty string.
        """
        self._name = val.strip()

    def clear(self) -> None:
        """*clear()* Reset all attributes to default values.

        Resets the entity's properties to their initial state.
        """
        # Reset parent class attributes
        super().clear()

        # Reset name and description attributes
        self.name = ""
        self.description = ""

    def __str__(self) -> str:
        """*__str__()* Detailed string representation with all attributes.

        Returns a comprehensive view of all non-private attributes, suitable for
        detailed inspection and logging.

        Returns:
            str: Detailed string representation with all attributes.
        """
        _attr_lt = []
        for attr, value in vars(self).items():
            # Skip private attributes starting with "__"
            if attr.startswith("__"):
                continue
            # Format callable attributes
            if callable(value):
                value = f"{value.__name__}{inspect.signature(value)}"
            # Format attribute name and value
            _attr_name = attr.lstrip("_")
            _attr_lt.append(f"{_attr_name}={repr(value)}")
        # Format the string with the class name and the attributes
        _str = f"{self.__class__.__name__}({', '.join(_attr_lt)})"
        return _str

    def __repr__(self) -> str:
        """*__repr__()* Detailed string representation.

        Returns:
            str: String representation.
        """
        return self.__str__()



# -*- coding: utf-8 -*-
"""
Module sllt.py
===========================================

Module for the custom **SingleLinkedList** data structure in *PyDASA*. Essential for Dimensional Analysis and Data Science operations.

Classes:
    **SingleLinkedList**: Implements a single linked list with methods for insertion, deletion, and traversal.

*IMPORTANT:* based on the implementations proposed by the following authors/books:

    #. Algorithms, 4th Edition, Robert Sedgewick and Kevin Wayne.
    #. Data Structure and Algorithms in Python, M.T. Goodrich, R. Tamassia, M.H. Goldwasser.
"""

# native python modules
# dataclass imports
from dataclasses import dataclass
# data type imports
from typing import List, Optional, Callable, Generic, Iterator, Any
# code inspection imports
import inspect

# custom modules
# linked list node implementation
from pydasa.structs.lists.ndlt import SLNode

# generic types and global variables
from pydasa.structs.types.generics import T
from pydasa.structs.types.generics import DFLT_DICT_KEY
from pydasa.structs.types.generics import VLD_IOTYPE_LT
from pydasa.structs.types.functions import dflt_cmp_function_lt
# generic error handling and type checking
from pydasa.validations.error import handle_error as error

# checking custom modules
assert T
assert DFLT_DICT_KEY
assert VLD_IOTYPE_LT
assert dflt_cmp_function_lt
assert error


@dataclass
class SingleLinkedList(Generic[T]):
    """**SingleLinkedList** implements a single linked list data structure for PyDASA.

    Args:
        Generic (T): Generic type for a Python data structure.

    Returns:
        SingleLinkedList: a generic single linked list data structure with the following attributes:
            - **cmp_function**: function to compare elements in the list.
            - **key**: key to identify the elements in the list.
            - **first**: reference to the first node of the list.
            - **last**: reference to the last node of the list.
            - **_size**: size of the list.
    """

    # the cmp_function is used to compare elements, not defined by default
    # :attr: cmp_function
    cmp_function: Optional[Callable[[T, T], int]] = None
    """
    Customizable comparison function for *SingleLinkedList* elements. Defaults to *dflt_cmp_function_lt()* from *PyDASA*, but can be overridden by the user.
    """

    # reference to the first node of the list
    # :attr: _first
    _first: Optional[SLNode[T]] = None
    """
    Reference to the first node of the *SingleLinkedList*.
    """

    # reference to the last node of the list
    # :attr: _last
    _last: Optional[SLNode[T]] = None
    """
    Reference to the last node of the *SingleLinkedList*.
    """

    # the key is used to compare elements, not defined by default
    # :attr: key
    key: Optional[str] = DFLT_DICT_KEY
    """
    Customizable key name for identifying elements in the *SingleLinkedList*. Defaults to *DFLT_DICT_KEY = '_id'* from *PyDASA*, but can be overridden by the user.
    """

    # by default, the list is empty
    # :attr: _size
    _size: int = 0
    """
    Size of the *SingleLinkedList*, starting at 0 and updated with each modification.
    """

    # input elements from python list
    # :attr: iodata
    iodata: Optional[List[T]] = None
    """
    Optional Python list for loading external data intho the *SingleLinkedList*. Defaults to *None* but can be provided during creation.
    """

    def __post_init__(self) -> None:
        """*__post_init__()* Initializes the *SingleLinkedList* after creation by setting attributes like *cmp_function*, *key*, *first*, *last*, and *iodata*.

        *NOTE:* Special method called automatically after object creation.
        """
        try:
            # if the key is not defined, use the default
            if self.key is None:
                self.key = DFLT_DICT_KEY

            # if the compare function is not defined, use the default
            if self.cmp_function is None:
                self.cmp_function = self.default_compare

            # if elements are provided, add them to the ArrayList
            if self.iodata is not None:
                if not isinstance(self.iodata, VLD_IOTYPE_LT):
                    raise TypeError(f"iodata must be a valid iterable type, got {type(self.iodata)}")

                for elm in self.iodata:
                    self.append(elm)

            # Clear iodata after processing
            self.iodata = None

        except Exception as err:
            self._error_handler(err)

    def default_compare(self, elm1: Any, elm2: Any) -> int:
        """*default_compare()* Default comparison function for *SingleLinkedList* elements. Compares two elements and returns:
            - 0 if they are equal,
            - 1 if the first is greater,
            - -1 if the first is smaller.

        Args:
            elm1 (Any): First element to compare.
            elm2 (Any): Second element to compare.

        Returns:
            int: Comparison result.
        """
        try:
            # default comparison needs the key to be defined
            if self.key is None:
                raise ValueError("Key must be set before comparison")
            return dflt_cmp_function_lt(elm1, elm2, self.key)
        except Exception as err:
            self._error_handler(err)
            raise   # Re-raise the exception after handling

    @property
    def size(self) -> int:
        """*size()* Property to retrieve the number of elements in the *SingleLinkedList*.

        Returns:
            int: number of elements in the *SingleLinkedList*.
        """
        return self._size

    @property
    def empty(self) -> bool:
        """*empty()* Property to check if the *SingleLinkedList* is empty.

        Returns:
            bool: True if the *SingleLinkedList* is empty, False otherwise.
        """
        return self._size == 0

    def clear(self) -> None:
        """*clear()* clears the *SingleLinkedList* by removing all elements and resetting the size to 0.

        NOTE: This method is used to empty the *SingleLinkedList* without deleting the object itself.
        """
        self._first = None
        self._last = None
        self._size = 0

    def prepend(self, elm: T) -> None:
        """*prepend()* adds an element to the beginning of the *SingleLinkedList*.

        Args:
            elm (T): element to be added to the beginning of the structure.
        """
        # if the element type is valid, add it to the list
        if self._validate_type(elm):
            # create a new node
            _new = SLNode(elm)
            _new.next = self._first
            self._first = _new
            if self.size == 0:
                self._last = self._first
            self._size += 1

    def append(self, elm: T) -> None:
        """*append()* adds an element to the end of the *SingleLinkedList*.

        Args:
            elm (T): element to be added to the end of the structure.
        """
        # if the element type is valid, add it to the list
        if self._validate_type(elm):
            # create a new node
            _new = SLNode(elm)
            if self._last is None:
                self._first = _new
            else:
                self._last.next = _new
            self._last = _new
            self._size += 1

    def insert(self, elm: T, pos: int) -> None:
        """*insert()* adds an element to the *SingleLinkedList* at a specific position.

        Args:
            elm (T): element to be added to the structure.
            pos (int): position where the element will be added.

        Raises:
            IndexError: error if the structure is empty.
            IndexError: error if the position is invalid.
            TypeError: error if the element type is invalid.
        """
        if not self.empty and self._validate_type(elm):
            if pos < 0 or pos > self.size:
                raise IndexError("Position is out of range")

            # create a new node
            _new = SLNode(elm)

            # if the position is the first, add it to the first
            if pos == 0:
                _new.next = self._first
                self._first = _new
                if self._last is None:  # Empty list case
                    self._last = _new

            # if the position is the last (append), add it to the end
            elif pos == self.size:
                if self._last is not None:
                    self._last.next = _new
                self._last = _new

            # otherwise, insert in the middle
            else:
                i = 0
                _cur = self._first
                _prev: Optional[SLNode[T]] = None
                while i < pos and _cur is not None:
                    _prev = _cur
                    _cur = _cur.next
                    i += 1
                if _prev is not None:
                    _new.next = _cur
                    _prev.next = _new
            # increment the size
            self._size += 1
        else:
            raise IndexError("Empty data structure")

    @property
    def first(self) -> T:
        """*first* Property to read the first element of the *SingleLinkedList*.

        Raises:
            IndexError: error if the structure is empty.

        Returns:
            T: the first element of the *SingleLinkedList*.
        """
        if self.empty or self._first is None:
            raise IndexError("Empty data structure")
        return self._first.data

    @property
    def last(self) -> T:
        """*last* Property to read the last element of the *SingleLinkedList*.

        Raises:
            Exception: error if the structure is empty.

        Returns:
             T: the last element of the *SingleLinkedList*.
        """
        if self.empty or self._last is None:
            raise IndexError("Empty data structure")
        return self._last.data

    def get(self, pos: int) -> T:
        """*get()* retrieves an element from the *SingleLinkedList* at a specific position.

        Args:
            pos (int): position of the element to be retrieved.

        Raises:
            IndexError: error if the structure is empty.
            IndexError: error if the position is invalid.

        Returns:
            T: the element at the specified position in the *SingleLinkedList*.
        """
        if self.empty:
            raise IndexError("Empty data structure")
        if pos < 0 or pos > self.size - 1:
            raise IndexError(f"Index {pos} is out of range")

        # current node starting at the first node
        _cur = self._first
        i = 0
        # iterate to the desired position
        while i != pos and _cur is not None:
            _cur = _cur.next
            i += 1

        if _cur is None:
            raise IndexError(f"Corrupted list structure at position {pos}")

        return _cur.data

    def __getitem__(self, pos: int) -> Optional[T]:
        """*__getitem__()* retrieves an element from the *SingleLinkedList* at a specific position.

        NOTE: This method is used to access the elements of the *SingleLinkedList* using the square brackets notation.

        Args:
            pos (int): position of the element to be retrieved.

        Raises:
            IndexError: error if the structure is empty.
            IndexError: error if the position is invalid.

        Returns:
            Optional[T]: the element at the specified position in the *SingleLinkedList*.
        """
        return self.get(pos)

    def pop_first(self) -> T:
        """*pop_first()* removes the first element from the *SingleLinkedList*.

        Raises:
            IndexError: error if the structure is empty.

        Returns:
            T: the first element of the *SingleLinkedList*.
        """
        # check if the list is empty
        if self.empty or self._first is None:
            raise IndexError("Empty data structure")

        # save the data before removing the node
        _data = self._first.data

        # move first pointer to the next node
        self._first = self._first.next
        self._size -= 1

        # if the list is now empty, set last to None
        if self._first is None:
            self._last = None

        return _data

    def pop_last(self) -> T:
        """*pop_last()* removes the last element from the *SingleLinkedList*.

        Raises:
            IndexError: error if the structure is empty.

        Returns:
            T: the last element of the *SingleLinkedList*.
        """
        # Check if the list is empty
        if self.empty or self._last is None:
            raise IndexError("Empty data structure")

        # Save the data before removing the node
        _data = self._last.data

        # if the list has only one element, set the first and last to None
        if self._first == self._last:
            self._first = None
            self._last = None

        # otherwise, remove the last element
        else:
            _cur = self._first
            # traverse the list to find the second-to-last element
            while _cur is not None and _cur.next != self._last:
                _cur = _cur.next

            # Ensure we found the second-to-last node
            if _cur is None:
                raise IndexError("Corrupted list structure")

            # rearrange the last element
            self._last = _cur
            self._last.next = None

        self._size -= 1
        return _data

    def remove(self, pos: int) -> T:
        """*remove()* removes an element from the *SingleLinkedList* at a specific position.

        Args:
            pos (int): position of the element to be removed.

        Raises:
            IndexError: error if the structure is empty.
            IndexError: error if the position is invalid.

        Returns:
            T: the element removed from the *SingleLinkedList*.
        """
        # check if the list is empty
        if self.empty or self._first is None:
            raise IndexError("Empty data structure")
        # check if the position is valid
        if pos < 0 or pos > self.size - 1:
            raise IndexError(f"Index {pos} is out of range")

        # if removing the first element
        if pos == 0:
            _data = self._first.data
            self._first = self._first.next
            # if list is now empty, update last
            if self._first is None:
                self._last = None
        # if removing from middle or end
        else:
            _cur = self._first
            _prev = self._first
            i = 0
            # traverse to the position
            while i != pos and _cur is not None:
                _prev = _cur
                _cur = _cur.next
                i += 1

            # Check if we found the node
            if _cur is None:
                raise IndexError(f"Corrupted list structure at position {pos}")

            _data = _cur.data
            _prev.next = _cur.next

            # if removing the last element, update last pointer
            if _cur == self._last:
                self._last = _prev

        self._size -= 1
        return _data

    def compare(self, elem1: T, elem2: T) -> int:
        """*compare()* compares two elements using the comparison function defined in the *SingleLinkedList*.

        Args:
            elem1 (T): first element to compare.
            elem2 (T): second element to compare.

        Raises:
            TypeError: error if the *cmp_function* is not defined.

        Returns:
            int: -1 if elem1 < elem2, 0 if elem1 == elem2, 1 if elem1 > elem2.
        """
        if self.cmp_function is None:
            # raise an exception if the cmp function is not defined
            raise TypeError("Undefined compare function!!!")
        # use the structure cmp function
        return self.cmp_function(elem1, elem2)

    def index_of(self, elm: T) -> int:
        """*index_of()* searches for the first occurrence of an element in the *SingleLinkedList*. If the element is found, it returns its index; otherwise, it returns -1.

        Args:
            elm (T): element to search for in the *SingleLinkedList*.

        Returns:
            int: index of the element in the *SingleLinkedList* or -1 if not found.
        """
        if self.empty:
            raise IndexError("Empty data structure")

        _idx = -1
        _node = self._first
        found = False
        i = 0

        # iterate through the list to find the element
        while not found and _node is not None and i < self.size:
            # using the structure cmp function
            if self.compare(elm, _node.data) == 0:
                found = True
                _idx = i
            else:
                _node = _node.next
                i += 1

        return _idx

    def update(self, new_data: T, pos: int) -> None:
        """*update()* updates an element in the *SingleLinkedList* at a specific position.

        Args:
            new_data (T): new data to be updated in the structure.
            pos (int): position of the element to be updated.

        Raises:
            IndexError: error if the structure is empty.
            IndexError: error if the position is invalid.
        """
        if self.empty:
            raise IndexError("Empty data structure")
        elif pos < 0 or pos > self.size - 1:
            raise IndexError(f"Index {pos} is out of range")
        # if the element type is valid, update the element
        elif self._validate_type(new_data):
            _cur = self._first
            i = 0
            while i != pos and _cur is not None:
                _cur = _cur.next
                i += 1

            if _cur is None:
                raise IndexError(f"Corrupted list structure at position {pos}")

            _cur.data = new_data

    def swap(self, pos1: int, pos2: int) -> None:
        """*swap()* swaps two elements in the *SingleLinkedList* at specific positions.


        Args:
            pos1 (int): position of the first element to swap.
            pos2 (int): position of the second element to swap.

        Raises:
            IndexError: error if the structure is empty.
            IndexError: error if the first position is invalid.
            IndexError: error if the second position is invalid.
        """
        if self.empty:
            raise IndexError("Empty data structure")
        elif pos1 < 0 or pos1 > self.size - 1:
            raise IndexError("Index", pos1, "is out of range")
        elif pos2 < 0 or pos2 > self.size - 1:
            raise IndexError("Index", pos2, "is out of range")
        info_pos1 = self.get(pos1)
        info_pos2 = self.get(pos2)
        self.update(info_pos2, pos1)
        self.update(info_pos1, pos2)

    def sublist(self, start: int, end: int) -> "SingleLinkedList[T]":
        """*sublist()* creates a new *SingleLinkedList* containing a sublist of elements from the original *SingleLinkedList*. The sublist is defined by the start and end indices.

        NOTE: The start index is inclusive, and the end index is inclusive.

        Args:
            start (int): start index of the sublist.
            end (int): end index of the sublist.

        Raises:
            IndexError: error if the structure is empty.
            IndexError: error if the start or end index are invalid.

        Returns:
            SingleLinkedList[T]: a new *SingleLinkedList* containing the sublist of elements.
        """
        if self.empty:
            raise IndexError("Empty data structure")
        elif start < 0 or end > self.size - 1 or start > end:
            raise IndexError(f"Invalid range: between [{start}, {end}]")

        sub_lt = SingleLinkedList(cmp_function=self.cmp_function,
                                  key=self.key)
        i = 0
        _cur = self._first

        while i != end + 1 and _cur is not None:
            if i >= start:
                sub_lt.append(_cur.data)
            _cur = _cur.next
            i += 1
        return sub_lt

    def concat(self, other: "SingleLinkedList[T]") -> "SingleLinkedList[T]":
        """*concat()* concatenates two *SingleLinkedList* objects. The elements of the second list are added to the end of the first list.

        NOTE: The *cmp_function* and *key* attributes of the two lists must be the same.

        Args:
            other (SingleLinkedList[T]): the second *SingleLinkedList* to be concatenated.

        Raises:
            TypeError: error if the *other* argument is not an *SingleLinkedList*.
            TypeError: error if the *key* attributes are not the same.
            TypeError: error if the *cmp_function* are not the same.

        Returns:
            SingleLinkedList[T]: the concatenated *SingleLinkedList* in the first list.
        """
        if not isinstance(other, SingleLinkedList):
            _msg = f"Structure is not an SingleLinkedList: {type(other)}"
            raise TypeError(_msg)
        if self.key != other.key:
            raise TypeError(f"Invalid key: {self.key} != {other.key}")
        # checking functional code of the cmp function
        if self.cmp_function is not None and other.cmp_function is not None:
            code1 = self.cmp_function.__code__.co_code
            code2 = other.cmp_function.__code__.co_code
            if code1 != code2:
                _msg = f"Invalid compare function: {self.cmp_function}"
                _msg += f" != {other.cmp_function}"
                raise TypeError(_msg)

        # Handle empty lists
        if other.empty:
            return self  # Nothing to concatenate

        # If self is empty, just copy other's strucure
        if self.empty:
            self._first = other._first
            self._last = other._last

        # concatenate the two lists
        elif self._last is not None:
            self._last.next = other._first
            self._last = other._last

        # update the size
        self._size = self.size + other.size
        return self

    def clone(self) -> "SingleLinkedList[T]":
        """*clone()* creates a copy of the *SingleLinkedList*. The new list is independent of the original list.

        NOTE: The elements of the new list are the same as the original list, but they are not references to the same objects.

        Returns:
            SingleLinkedList[T]: a new *SingleLinkedList* with the same elements as the original list.
        """
        # create new list
        copy_lt = SingleLinkedList(cmp_function=self.cmp_function,
                                   key=self.key)
        # get the first node of the original list
        _cur = self._first
        # traverse the list and add the elements to the new list
        while _cur is not None:
            copy_lt.append(_cur.data)
            _cur = _cur.next
        return copy_lt

    def _error_handler(self, err: Exception) -> None:
        """*_error_handler()* to process the context (package/class), function name (method), and the error (exception) that was raised to format a detailed error message and traceback.

        Args:
            err (Exception): Python raised exception.
        """
        _context = self.__class__.__name__
        _function_name = "unknown"
        frame = inspect.currentframe()
        if frame is not None:
            if frame.f_back is not None:
                _function_name = frame.f_back.f_code.co_name
        else:
            _function_name = "unknown"
        error(_context, _function_name, err)

    def _validate_type(self, elm: T) -> bool:
        """*_validate_type()* checks if the type of the element is valid. If the structure is empty, the type is valid. If the structure is not empty, the type must be the same as the first element in the list.
        This is used to check the type of the element before adding it to the list.

        Args:
            elm (T): element to be added to the structure.

        Raises:
            TypeError: error if the type of the element is not valid.

        Returns:
            bool: True if the type is valid, False otherwise.
        """
        # if the structure is not empty, check the first element type
        if not self.empty and self._first is not None:
            # raise an exception if the type is not valid at the first element
            if not isinstance(elm, type(self._first.data)):
                _msg = f"Invalid data type: {type(elm)} "
                _msg += f"!= {type(self._first.data)}"
                raise TypeError(_msg)
        # otherwise, any type is valid
        return True

    def __iter__(self) -> Iterator[T]:
        """*__iter__()* to iterate over the elements of the *SingleLinkedList*. This method returns an iterator object that can be used to iterate over the elements of the list.

        NOTE: This is used to iterate over the nodes of the list using a for loop.

        Returns:
            Iterator[T]: an iterator object that can be used to iterate over the nodes of the list.
        """
        try:
            # TODO do I need the try/except block?
            _cur = self._first
            while _cur is not None:
                yield _cur.data
                _cur = _cur.next
        except Exception as err:
            self._error_handler(err)

    def __len__(self) -> int:
        """*__len__()* to get the number of elements in the *SingleLinkedList*. This method returns the size of the list.

        Returns:
            int: the number of elements in the *SingleLinkedList*.
        """
        return self._size

    def __str__(self) -> str:
        """*__str__()* to get the string representation of the *SingleLinkedList*. This method returns a string with the elements of the list separated by commas.

        Returns:
            str: string representation of the *SingleLinkedList*.
        """
        _attr_lt = []
        for attr, value in vars(self).items():
            # Skip private attributes starting with "__"
            if attr.startswith("__"):
                continue
            # Format callable attributes
            if callable(value):
                value = f"{value.__name__}{inspect.signature(value)}"
            # Format attribute name and value
            _attr_name = attr.lstrip("_")
            _attr_lt.append(f"{_attr_name}={repr(value)}")
        # format the string with the SingleLinkedList class name and the attributes
        _str = f"{self.__class__.__name__}({', '.join(_attr_lt)})"
        return _str

    def __repr__(self) -> str:
        """*__repr__()* get the string representation of the *SingleLinkedList*. This method returns a string representation,

        Returns:
            str: string representation of the *SingleLinkedList*.
        """
        return self.__str__()



import os
import re
from setuptools import setup, find_packages


def read_version() -> str:
    """read the project version in a preset file.

    Raises:
        RuntimeError: if the version string is not found.

    Returns:
        str: project version.
    """
    version_file = os.path.join("src", "pydasa", "_version.py")
    with open(version_file, "r", encoding="utf-8") as f:
        content = f.read()
    m = re.search(r'__version__\s*=\s*[\'"]([^\'"]+)[\'"]', content)
    if not m:
        raise RuntimeError("Unable to find version string.")
    return m.group(1)


def read_long_description() -> str:
    """Read the long description from the README file.

    Returns:
        str: project long description.
    """
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
    return long_description


setup(
    name="pydasa",
    version=read_version(),
    description="Python package for Dimensional Analysis for Scientific Applications and Software Architecture (PyDASA).",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    long_description=read_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/DASA-Design/PyDASA",
    author="@SFAM",
    author_email="sa-artea@uniandes.edu.co",
    # NOTE keep check classifiers!!!
    # license="GPL-3.0",
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
    ],
    # install_requires=[
    #     "antlr4-python3-runtime==4.11",
    #     "numpy>=1.26.4",
    #     "scipy>=1.13.0",
    #     "sympy>=1.12",
    #     "matplotlib>=3.8.0",
    #     "pandas>=2.1.0",
    #     "SALib>=1.4.5",
    # ],
    # extras_require={
    #     "dev": [
    #         "pytest>=8.1.1",
    #         "twine>=6.1.0",
    #     ],
    #     "docs": [
    #         "sphinx>=7.3.7",
    #         "sphinx-rtd-theme>=1.3.0",
    #     ],
    # },
    python_requires=">=3.10",
)
